@@ -1,4 +1,30 @@
-#include "fitz-imp.h"
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
+#include "mupdf/fitz.h"
+
+#include "context-imp.h"
+#include "image-imp.h"
+#include "pixmap-imp.h"
 
 #include <string.h>
 #include <math.h>
@@ -16,27 +42,25 @@ fz_key_storable_needs_reaping(fz_context *ctx, const fz_key_storable *ks)
 
 #define SCALABLE_IMAGE_DPI 96
 
-struct fz_compressed_image_s
+struct fz_compressed_image
 {
 	fz_image super;
-	fz_pixmap *tile;
 	fz_compressed_buffer *buffer;
 };
 
-struct fz_pixmap_image_s
+struct fz_pixmap_image
 {
 	fz_image super;
 	fz_pixmap *tile;
 };
 
-typedef struct fz_image_key_s fz_image_key;
-
-struct fz_image_key_s {
+typedef struct
+{
 	int refs;
 	fz_image *image;
 	int l2factor;
 	fz_irect rect;
-};
+} fz_image_key;
 
 fz_image *
 fz_keep_image(fz_context *ctx, fz_image *image)
@@ -93,7 +117,7 @@ fz_cmp_image_key(fz_context *ctx, void *k0_, void *k1_)
 }
 
 static void
-fz_format_image_key(fz_context *ctx, char *s, int n, void *key_)
+fz_format_image_key(fz_context *ctx, char *s, size_t n, void *key_)
 {
 	fz_image_key *key = (fz_image_key *)key_;
 	fz_snprintf(s, n, "(image %d x %d sf=%d)", key->image->w, key->image->h, key->l2factor);
@@ -109,6 +133,7 @@ fz_needs_reap_image_key(fz_context *ctx, void *key_)
 
 static const fz_store_type fz_image_store_type =
 {
+	"fz_image",
 	fz_make_hash_image_key,
 	fz_keep_image_key,
 	fz_drop_image_key,
@@ -124,15 +149,44 @@ fz_drop_image(fz_context *ctx, fz_image *image)
 }
 
 static void
-fz_mask_color_key(fz_pixmap *pix, int n, const int *colorkey)
+fz_mask_color_key(fz_pixmap *pix, int n, int bpc, const int *colorkey)
 {
 	unsigned char *p = pix->samples;
 	int w;
 	int k, t;
 	int h = pix->h;
-	int stride = pix->stride - pix->w * pix->n;
+	size_t stride = pix->stride - pix->w * (size_t)pix->n;
+	int scaledcolorkey[FZ_MAX_COLORS * 2];
+	int scale, shift, max;
+
 	if (pix->w == 0)
 		return;
+
+	for (k = 0; k < 2 * n; k++)
+		scaledcolorkey[k] = fz_clampi(colorkey[k], 0, (1 << bpc) - 1);
+
+	switch (bpc)
+	{
+	case 1: scale = 255; shift = 0; max = 1; break;
+	case 2: scale = 85; shift = 0; max = 3; break;
+	case 4: scale = 17; shift = 0; max = 15; break;
+	default:
+	case 8: scale = 1; shift = 0; max = 0xff; break;
+	case 16: scale = 1; shift = 8; max = 0xffff; break;
+	case 24: scale = 1; shift = 16; max = 0xffffff; break;
+	case 32: scale = 1; shift = 24; max = 0xffffffff; break;
+	}
+
+	for (k = 0; k < 2 * n; k++)
+		scaledcolorkey[k] = fz_clampi(colorkey[k], 0, max);
+
+	if (scale > 1)
+		for (k = 0; k < 2 * n; k++)
+			scaledcolorkey[k] *= scale;
+	else if (shift > 0)
+		for (k = 0; k < 2 * n; k++)
+			scaledcolorkey[k] >>= shift;
+
 	while (h--)
 	{
 		w = pix->w;
@@ -140,7 +194,7 @@ fz_mask_color_key(fz_pixmap *pix, int n, const int *colorkey)
 		{
 			t = 1;
 			for (k = 0; k < n; k++)
-				if (p[k] < colorkey[k * 2] || p[k] > colorkey[k * 2 + 1])
+				if (p[k] < scaledcolorkey[k * 2] || p[k] > scaledcolorkey[k * 2 + 1])
 					t = 0;
 			if (t)
 				for (k = 0; k < pix->n; k++)
@@ -153,29 +207,47 @@ fz_mask_color_key(fz_pixmap *pix, int n, const int *colorkey)
 }
 
 static void
-fz_unblend_masked_tile(fz_context *ctx, fz_pixmap *tile, fz_image *image)
+fz_unblend_masked_tile(fz_context *ctx, fz_pixmap *tile, fz_image *image, const fz_irect *isa)
 {
-	fz_pixmap *mask = fz_get_pixmap_from_image(ctx, image->mask, NULL, NULL, NULL, NULL);
-	unsigned char *s = mask->samples;
-	unsigned char *d = tile->samples;
+	fz_pixmap *mask;
+	unsigned char *s, *d = tile->samples;
 	int n = tile->n;
 	int k;
-	int sstride = mask->stride - mask->w * mask->n;
-	int dstride = tile->stride - tile->w * tile->n;
-	int h = mask->h;
+	size_t sstride, dstride = tile->stride - tile->w * (size_t)tile->n;
+	int h;
+	fz_irect subarea;
 
-	if (tile->w != mask->w || tile->h != mask->h)
+	/* We need at least as much of the mask as there was of the tile. */
+	if (isa)
+		subarea = *isa;
+	else
 	{
-		fz_warn(ctx, "mask must be of same size as image for /Matte");
-		fz_drop_pixmap(ctx, mask);
-		return;
+		subarea.x0 = 0;
+		subarea.y0 = 0;
+		subarea.x1 = tile->w;
+		subarea.y1 = tile->h;
 	}
 
-	if (mask->w != 0)
+	mask = fz_get_pixmap_from_image(ctx, image->mask, &subarea, NULL, NULL, NULL);
+	s = mask->samples;
+	/* RJW: Urgh, bit of nastiness here. fz_pixmap_from_image will either return
+	 * an exact match for the subarea we asked for, or the full image, and the
+	 * normal way to know is that the matrix will be updated. That doesn't help
+	 * us here. */
+	if (image->mask->w == mask->w && image->mask->h == mask->h) {
+		subarea.x0 = 0;
+		subarea.y0 = 0;
+	}
+	if (isa)
+		s += (isa->x0 - subarea.x0) * (size_t)mask->n + (isa->y0 - subarea.y0) * (size_t)mask->stride;
+	sstride = mask->stride - tile->w * (size_t)mask->n;
+	h = tile->h;
+
+	if (tile->w != 0)
 	{
 		while (h--)
 		{
-			int w = mask->w;
+			int w = tile->w;
 			do
 			{
 				if (*s == 0)
@@ -305,8 +377,181 @@ static void fz_compute_image_key(fz_context *ctx, fz_image *image, fz_matrix *ct
 		key->l2factor = 0;
 }
 
+typedef struct {
+	fz_stream *src;
+	size_t l_skip; /* Number of bytes to skip on the left. */
+	size_t r_skip; /* Number of bytes to skip on the right. */
+	size_t b_skip; /* Number of bytes to skip on the bottom. */
+	int lines; /* Number of lines left to copy. */
+	size_t stride; /* Number of bytes to read in the image. */
+	size_t nskip; /* Number of bytes left to skip on this line. */
+	size_t nread; /* Number of bytes left to read on this line. */
+} subarea_state;
+
+static int
+subarea_next(fz_context *ctx, fz_stream *stm, size_t len)
+{
+	subarea_state *state = stm->state;
+	size_t n;
+
+	stm->wp = stm->rp = NULL;
+
+	while (state->nskip > 0)
+	{
+		n = fz_skip(ctx, state->src, state->nskip);
+		if (n == 0)
+			return EOF;
+		state->nskip -= n;
+	}
+	if (state->lines == 0)
+		return EOF;
+	n = fz_available(ctx, state->src, state->nread);
+	if (n > state->nread)
+		n = state->nread;
+	if (n == 0)
+		return EOF;
+	stm->rp = state->src->rp;
+	stm->wp = stm->rp + n;
+	stm->pos += n;
+	state->src->rp = stm->wp;
+	state->nread -= n;
+	if (state->nread == 0)
+	{
+		state->lines--;
+		if (state->lines == 0)
+			state->nskip = state->r_skip + state->b_skip;
+		else
+			state->nskip = state->l_skip + state->r_skip;
+		state->nread = state->stride;
+	}
+	return *stm->rp++;
+}
+
+static void
+subarea_drop(fz_context *ctx, void *state)
+{
+	fz_free(ctx, state);
+}
+
+static fz_stream *
+subarea_stream(fz_context *ctx, fz_stream *stm, fz_image *image, const fz_irect *subarea, int l2factor)
+{
+	subarea_state *state;
+	int f = 1<<l2factor;
+	int stream_w = (image->w + f - 1)>>l2factor;
+	size_t stream_stride = (stream_w * (size_t)image->n * image->bpc + 7) / 8;
+	int l_margin = subarea->x0 >> l2factor;
+	int t_margin = subarea->y0 >> l2factor;
+	int r_margin = (image->w + f - 1 - subarea->x1) >> l2factor;
+	int b_margin = (image->h + f - 1 - subarea->y1) >> l2factor;
+	size_t l_skip = (l_margin * (size_t)image->n * image->bpc)/8;
+	size_t r_skip = (r_margin * (size_t)image->n * image->bpc + 7)/8;
+	size_t t_skip = t_margin * stream_stride;
+	size_t b_skip = b_margin * stream_stride;
+	int h = (subarea->y1 - subarea->y0 + f - 1) >> l2factor;
+	int w = (subarea->x1 - subarea->x0 + f - 1) >> l2factor;
+	size_t stride = (w * (size_t)image->n * image->bpc + 7) / 8;
+
+	state = fz_malloc_struct(ctx, subarea_state);
+	state->src = stm;
+	state->l_skip = l_skip;
+	state->r_skip = r_skip;
+	state->b_skip = b_skip;
+	state->lines = h;
+	state->nskip = l_skip+t_skip;
+	state->stride = stride;
+	state->nread = stride;
+
+	return fz_new_stream(ctx, state, subarea_next, subarea_drop);
+}
+
+typedef struct
+{
+	fz_stream *src;
+	int w; /* Width in source pixels. */
+	int h; /* Height (remaining) in scanlines. */
+	int n; /* Number of components. */
+	int f; /* Fill level (how many scanlines we've copied in). */
+	size_t r; /* How many samples Remain to be filled in this line. */
+	int l2; /* The amount of subsampling we're doing. */
+	unsigned char data[1];
+} l2sub_state;
+
+static void
+subsample_drop(fz_context *ctx, void *state)
+{
+	fz_free(ctx, state);
+}
+
+static int
+subsample_next(fz_context *ctx, fz_stream *stm, size_t len)
+{
+	l2sub_state *state = (l2sub_state *)stm->state;
+	size_t fill;
+
+	stm->rp = stm->wp = &state->data[0];
+	if (state->h == 0)
+		return EOF;
+
+	/* Copy in data */
+	do
+	{
+		if (state->r == 0)
+			state->r = state->w * (size_t)state->n;
+
+		while (state->r > 0)
+		{
+			size_t a;
+			a = fz_available(ctx, state->src, state->r);
+			if (a == 0)
+				return EOF;
+			if (a > state->r)
+				a = state->r;
+			memcpy(&state->data[state->w * (size_t)state->n * (state->f+1) - state->r],
+				state->src->rp, a);
+			state->src->rp += a;
+			state->r -= a;
+		}
+		state->f++;
+		state->h--;
+	}
+	while (state->h > 0 && state->f != (1<<state->l2));
+
+	/* Perform the subsample */
+	fz_subsample_pixblock(state->data, state->w, state->f, state->n, state->l2, state->w * (size_t)state->n);
+	state->f = 0;
+
+	/* Update data pointers. */
+	fill = ((state->w + (1<<state->l2) - 1)>>state->l2) * (size_t)state->n;
+	stm->pos += fill;
+	stm->rp = &state->data[0];
+	stm->wp = &state->data[fill];
+
+	return *stm->rp++;
+}
+
+static fz_stream *
+subsample_stream(fz_context *ctx, fz_stream *src, int w, int h, int n, int l2extra)
+{
+	l2sub_state *state = fz_malloc(ctx, sizeof(l2sub_state) + w*(size_t)(n<<l2extra));
+
+	state->src = src;
+	state->w = w;
+	state->h = h;
+	state->n = n;
+	state->f = 0;
+	state->r = 0;
+	state->l2 = l2extra;
+
+	return fz_new_stream(ctx, state, subsample_next, subsample_drop);
+}
+
+/* l2factor is the amount of subsampling that the decoder is going to be
+ * doing for us already. (So for JPEG 0,1,2,3 corresponding to 1, 2, 4,
+ * 8. For other formats, probably 0.). l2extra is the additional amount
+ * of subsampling we should perform here. */
 fz_pixmap *
-fz_decomp_image_from_stream(fz_context *ctx, fz_stream *stm, fz_compressed_image *cimg, fz_irect *subarea, int indexed, int l2factor)
+fz_decomp_image_from_stream(fz_context *ctx, fz_stream *stm, fz_compressed_image *cimg, fz_irect *subarea, int indexed, int l2factor, int *l2extra)
 {
 	fz_image *image = &cimg->super;
 	fz_pixmap *tile = NULL;
@@ -315,74 +560,71 @@ fz_decomp_image_from_stream(fz_context *ctx, fz_stream *stm, fz_compressed_image
 	int f = 1<<l2factor;
 	int w = image->w;
 	int h = image->h;
+	int matte = image->use_colorkey && image->mask;
+	fz_stream *read_stream = stm;
+	fz_stream *sstream = NULL;
+	fz_stream *l2stream = NULL;
+	fz_stream *unpstream = NULL;
 
+	if (matte)
+	{
+		/* Can't do l2factor decoding */
+		if (image->w != image->mask->w || image->h != image->mask->h)
+		{
+			fz_warn(ctx, "mask must be of same size as image for /Matte");
+			matte = 0;
+		}
+		assert(l2factor == 0);
+	}
 	if (subarea)
 	{
-		fz_adjust_image_subarea(ctx, image, subarea, l2factor);
-		w = (subarea->x1 - subarea->x0);
-		h = (subarea->y1 - subarea->y0);
+		if (subarea->x0 == 0 && subarea->x1 == image->w &&
+			subarea->y0 == 0 && subarea->y1 == image->h)
+			subarea = NULL;
+		else
+		{
+			fz_adjust_image_subarea(ctx, image, subarea, l2factor);
+			w = (subarea->x1 - subarea->x0);
+			h = (subarea->y1 - subarea->y0);
+		}
 	}
 	w = (w + f - 1) >> l2factor;
 	h = (h + f - 1) >> l2factor;
 
 	fz_var(tile);
 	fz_var(samples);
+	fz_var(sstream);
+	fz_var(unpstream);
+	fz_var(l2stream);
 
 	fz_try(ctx)
 	{
 		int alpha = (image->colorspace == NULL);
 		if (image->use_colorkey)
 			alpha = 1;
+
+		if (subarea)
+			read_stream = sstream = subarea_stream(ctx, stm, image, subarea, l2factor);
+		if (image->bpc != 8 || image->use_colorkey)
+			read_stream = unpstream = fz_unpack_stream(ctx, read_stream, image->bpc, w, h, image->n, indexed, image->use_colorkey, 0);
+		if (l2extra && *l2extra && !indexed)
+		{
+			read_stream = l2stream = subsample_stream(ctx, read_stream, w, h, image->n + image->use_colorkey, *l2extra);
+			w = (w + (1<<*l2extra) - 1)>>*l2extra;
+			h = (h + (1<<*l2extra) - 1)>>*l2extra;
+			*l2extra = 0;
+		}
+
 		tile = fz_new_pixmap(ctx, image->colorspace, w, h, NULL, alpha);
 		if (image->interpolate & FZ_PIXMAP_FLAG_INTERPOLATE)
 			tile->flags |= FZ_PIXMAP_FLAG_INTERPOLATE;
 		else
 			tile->flags &= ~FZ_PIXMAP_FLAG_INTERPOLATE;
 
-		stride = (w * image->n * image->bpc + 7) / 8;
+		samples = tile->samples;
+		stride = tile->stride;
 
-		samples = fz_malloc_array(ctx, h, stride);
-
-		if (subarea)
-		{
-			int hh;
-			unsigned char *s = samples;
-			int stream_w = (image->w + f - 1)>>l2factor;
-			size_t stream_stride = (stream_w * image->n * image->bpc + 7) / 8;
-			int l_margin = subarea->x0 >> l2factor;
-			int t_margin = subarea->y0 >> l2factor;
-			int r_margin = (image->w + f - 1 - subarea->x1) >> l2factor;
-			int b_margin = (image->h + f - 1 - subarea->y1) >> l2factor;
-			int l_skip = (l_margin * image->n * image->bpc)/8;
-			int r_skip = (r_margin * image->n * image->bpc + 7)/8;
-			size_t t_skip = t_margin * stream_stride + l_skip;
-			size_t b_skip = b_margin * stream_stride + r_skip;
-			size_t l = fz_skip(ctx, stm, t_skip);
-			len = 0;
-			if (l == t_skip)
-			{
-				hh = h;
-				do
-				{
-					l = fz_read(ctx, stm, s, stride);
-					s += l;
-					len += l;
-					if (l < stride)
-						break;
-					if (--hh == 0)
-						break;
-					l = fz_skip(ctx, stm, r_skip + l_skip);
-					if (l < (size_t)(r_skip + l_skip))
-						break;
-				}
-				while (1);
-				(void)fz_skip(ctx, stm, r_skip + b_skip);
-			}
-		}
-		else
-		{
-			len = fz_read(ctx, stm, samples, h * stride);
-		}
+		len = fz_read(ctx, read_stream, samples, h * stride);
 
 		/* Pad truncated images */
 		if (len < stride * h)
@@ -401,20 +643,15 @@ fz_decomp_image_from_stream(fz_context *ctx, fz_stream *stm, fz_compressed_image
 				p[i] = ~p[i];
 		}
 
-		fz_unpack_tile(ctx, tile, samples, image->n, image->bpc, stride, indexed);
-
-		fz_free(ctx, samples);
-		samples = NULL;
-
 		/* color keyed transparency */
 		if (image->use_colorkey && !image->mask)
-			fz_mask_color_key(tile, image->n, image->colorkey);
+			fz_mask_color_key(tile, image->n, image->bpc, image->colorkey);
 
 		if (indexed)
 		{
 			fz_pixmap *conv;
 			fz_decode_indexed_tile(ctx, tile, image->decode, (1 << image->bpc) - 1);
-			conv = fz_expand_indexed_pixmap(ctx, tile, alpha);
+			conv = fz_convert_indexed_pixmap_to_base(ctx, tile);
 			fz_drop_pixmap(ctx, tile);
 			tile = conv;
 		}
@@ -424,13 +661,18 @@ fz_decomp_image_from_stream(fz_context *ctx, fz_stream *stm, fz_compressed_image
 		}
 
 		/* pre-blended matte color */
-		if (image->use_colorkey && image->mask)
-			fz_unblend_masked_tile(ctx, tile, image);
+		if (matte)
+			fz_unblend_masked_tile(ctx, tile, image, subarea);
+	}
+	fz_always(ctx)
+	{
+		fz_drop_stream(ctx, sstream);
+		fz_drop_stream(ctx, unpstream);
+		fz_drop_stream(ctx, l2stream);
 	}
 	fz_catch(ctx)
 	{
 		fz_drop_pixmap(ctx, tile);
-		fz_free(ctx, samples);
 		fz_rethrow(ctx);
 	}
 
@@ -459,7 +701,6 @@ drop_compressed_image(fz_context *ctx, fz_image *image_)
 {
 	fz_compressed_image *image = (fz_compressed_image *)image_;
 
-	fz_drop_pixmap(ctx, image->tile);
 	fz_drop_compressed_buffer(ctx, image->buffer);
 }
 
@@ -480,6 +721,16 @@ compressed_image_get_pixmap(fz_context *ctx, fz_image *image_, fz_irect *subarea
 	int indexed;
 	fz_pixmap *tile;
 	int can_sub = 0;
+	int local_l2factor;
+
+	/* If we are using matte, then the decode code requires both image and tile sizes
+	 * to match. The simplest way to ensure this is to do no native l2factor decoding.
+	 */
+	if (image->super.use_colorkey && image->super.mask)
+	{
+		local_l2factor = 0;
+		l2factor = &local_l2factor;
+	}
 
 	/* We need to make a new one. */
 	/* First check for ones that we can't decode using streams */
@@ -504,7 +755,7 @@ compressed_image_get_pixmap(fz_context *ctx, fz_image *image_, fz_irect *subarea
 		tile = fz_load_jxr(ctx, image->buffer->buffer->data, image->buffer->buffer->len);
 		break;
 	case FZ_IMAGE_JPX:
-		tile = fz_load_jpx(ctx, image->buffer->buffer->data, image->buffer->buffer->len, NULL);
+		tile = fz_load_jpx(ctx, image->buffer->buffer->data, image->buffer->buffer->len, image->super.colorspace);
 		break;
 	case FZ_IMAGE_JPEG:
 		/* Scan JPEG stream and patch missing height values in header */
@@ -512,7 +763,7 @@ compressed_image_get_pixmap(fz_context *ctx, fz_image *image_, fz_irect *subarea
 			unsigned char *s = image->buffer->buffer->data;
 			unsigned char *e = s + image->buffer->buffer->len;
 			unsigned char *d;
-			for (d = s + 2; s < d && d < e - 9 && d[0] == 0xFF; d += (d[2] << 8 | d[3]) + 2)
+			for (d = s + 2; s < d && d + 9 < e && d[0] == 0xFF; d += (d[2] << 8 | d[3]) + 2)
 			{
 				if (d[1] < 0xC0 || (0xC3 < d[1] && d[1] < 0xC9) || 0xCB < d[1])
 					continue;
@@ -534,22 +785,13 @@ compressed_image_get_pixmap(fz_context *ctx, fz_image *image_, fz_irect *subarea
 				native_l2factor -= *l2factor;
 			indexed = fz_colorspace_is_indexed(ctx, image->super.colorspace);
 			can_sub = 1;
-			tile = fz_decomp_image_from_stream(ctx, stm, image, subarea, indexed, native_l2factor);
+			tile = fz_decomp_image_from_stream(ctx, stm, image, subarea, indexed, native_l2factor, l2factor);
 		}
 		fz_always(ctx)
 			fz_drop_stream(ctx, stm);
 		fz_catch(ctx)
 			fz_rethrow(ctx);
 
-		/* CMYK JPEGs in XPS documents have to be inverted */
-		if (image->super.invert_cmyk_jpeg &&
-			image->buffer->params.type == FZ_IMAGE_JPEG &&
-			fz_colorspace_is_cmyk(ctx, image->super.colorspace) &&
-			image->buffer->params.u.jpeg.color_transform)
-		{
-			fz_invert_pixmap(ctx, tile);
-		}
-
 		break;
 	}
 
@@ -581,7 +823,7 @@ update_ctm_for_subarea(fz_matrix *ctm, const fz_irect *subarea, int w, int h)
 {
 	fz_matrix m;
 
-	if (subarea->x0 == 0 && subarea->y0 == 0 && subarea->x1 == w && subarea->y1 == h)
+	if (ctm == NULL || (subarea->x0 == 0 && subarea->y0 == 0 && subarea->x1 == w && subarea->y1 == h))
 		return;
 
 	m.a = (float) (subarea->x1 - subarea->x0) / w;
@@ -644,10 +886,12 @@ fz_get_pixmap_from_image(fz_context *ctx, fz_image *image, const fz_irect *subar
 	fz_pixmap *tile;
 	int l2factor, l2factor_remaining;
 	fz_image_key key;
-	fz_image_key *keyp;
+	fz_image_key *keyp = NULL;
 	int w;
 	int h;
 
+	fz_var(keyp);
+
 	if (!image)
 		return NULL;
 
@@ -741,16 +985,19 @@ fz_get_pixmap_from_image(fz_context *ctx, fz_image *image, const fz_irect *subar
 		}
 	}
 
-	/* Now we try to cache the pixmap. Any failure here will just result
-	 * in us not caching. */
-	keyp = fz_malloc_struct(ctx, fz_image_key);
-	keyp->refs = 1;
-	keyp->image = fz_keep_image_store_key(ctx, image);
-	keyp->l2factor = l2factor;
-	keyp->rect = key.rect;
 	fz_try(ctx)
 	{
-		fz_pixmap *existing_tile = fz_store_item(ctx, keyp, tile, fz_pixmap_size(ctx, tile), &fz_image_store_type);
+		fz_pixmap *existing_tile;
+
+		/* Now we try to cache the pixmap. Any failure here will just result
+		 * in us not caching. */
+		keyp = fz_malloc_struct(ctx, fz_image_key);
+		keyp->refs = 1;
+		keyp->image = fz_keep_image_store_key(ctx, image);
+		keyp->l2factor = l2factor;
+		keyp->rect = key.rect;
+
+		existing_tile = fz_store_item(ctx, keyp, tile, fz_pixmap_size(ctx, tile), &fz_image_store_type);
 		if (existing_tile)
 		{
 			/* We already have a tile. This must have been produced by a
@@ -771,6 +1018,12 @@ fz_get_pixmap_from_image(fz_context *ctx, fz_image *image, const fz_irect *subar
 	return tile;
 }
 
+fz_pixmap *
+fz_get_unscaled_pixmap_from_image(fz_context *ctx, fz_image *image)
+{
+	return fz_get_pixmap_from_image(ctx, image, NULL /*subarea*/, NULL /*ctm*/, NULL /*dw*/, NULL /*dh*/);
+}
+
 static size_t
 pixmap_image_get_size(fz_context *ctx, fz_image *image)
 {
@@ -810,7 +1063,7 @@ fz_new_image_from_pixmap(fz_context *ctx, fz_pixmap *pixmap, fz_image *mask)
 fz_image *
 fz_new_image_of_size(fz_context *ctx, int w, int h, int bpc, fz_colorspace *colorspace,
 		int xres, int yres, int interpolate, int imagemask, float *decode,
-		int *colorkey, fz_image *mask, int size,
+		int *colorkey, fz_image *mask, size_t size,
 		fz_image_get_pixmap_fn *get_pixmap,
 		fz_image_get_size_fn *get_size,
 		fz_drop_image_fn *drop)
@@ -833,7 +1086,6 @@ fz_new_image_of_size(fz_context *ctx, int w, int h, int bpc, fz_colorspace *colo
 	image->bpc = bpc;
 	image->n = (colorspace ? fz_colorspace_n(ctx, colorspace) : 1);
 	image->colorspace = fz_keep_colorspace(ctx, colorspace);
-	image->invert_cmyk_jpeg = 1;
 	image->interpolate = interpolate;
 	image->imagemask = imagemask;
 	image->use_colorkey = (colorkey != NULL);
@@ -886,7 +1138,7 @@ compressed_image_get_size(fz_context *ctx, fz_image *image)
 	if (image == NULL)
 		return 0;
 
-	return sizeof(fz_pixmap_image) + fz_pixmap_size(ctx, im->tile) + (im->buffer && im->buffer->buffer ? im->buffer->buffer->cap : 0);
+	return sizeof(fz_pixmap_image) + (im->buffer && im->buffer->buffer ? im->buffer->buffer->cap : 0);
 }
 
 fz_image *
@@ -930,20 +1182,6 @@ void fz_set_compressed_image_buffer(fz_context *ctx, fz_compressed_image *image,
 	((fz_compressed_image *)image)->buffer = buf; /* Note: compressed buffers are not reference counted */
 }
 
-fz_pixmap *fz_compressed_image_tile(fz_context *ctx, fz_compressed_image *image)
-{
-	if (image == NULL || image->super.get_pixmap != compressed_image_get_pixmap)
-		return NULL;
-	return ((fz_compressed_image *)image)->tile;
-}
-
-void fz_set_compressed_image_tile(fz_context *ctx, fz_compressed_image *image, fz_pixmap *pix)
-{
-	assert(image != NULL && image->super.get_pixmap == compressed_image_get_pixmap);
-	fz_drop_pixmap(ctx, ((fz_compressed_image *)image)->tile);
-	((fz_compressed_image *)image)->tile = fz_keep_pixmap(ctx, pix);
-}
-
 fz_pixmap *fz_pixmap_image_tile(fz_context *ctx, fz_pixmap_image *image)
 {
 	if (image == NULL || image->super.get_pixmap != pixmap_image_get_pixmap)
@@ -957,11 +1195,58 @@ void fz_set_pixmap_image_tile(fz_context *ctx, fz_pixmap_image *image, fz_pixmap
 	((fz_pixmap_image *)image)->tile = pix;
 }
 
+const char *
+fz_image_type_name(int type)
+{
+	switch (type)
+	{
+	default:
+	case FZ_IMAGE_UNKNOWN: return "unknown";
+	case FZ_IMAGE_RAW: return "raw";
+	case FZ_IMAGE_FAX: return "fax";
+	case FZ_IMAGE_FLATE: return "flate";
+	case FZ_IMAGE_LZW: return "lzw";
+	case FZ_IMAGE_RLD: return "rld";
+	case FZ_IMAGE_BMP: return "bmp";
+	case FZ_IMAGE_GIF: return "gif";
+	case FZ_IMAGE_JBIG2: return "jbig2";
+	case FZ_IMAGE_JPEG: return "jpeg";
+	case FZ_IMAGE_JPX: return "jpx";
+	case FZ_IMAGE_JXR: return "jxr";
+	case FZ_IMAGE_PNG: return "png";
+	case FZ_IMAGE_PNM: return "pnm";
+	case FZ_IMAGE_TIFF: return "tiff";
+	}
+}
+
+int
+fz_lookup_image_type(const char *type)
+{
+	if (type == NULL) return FZ_IMAGE_UNKNOWN;
+	if (!strcmp(type, "raw")) return FZ_IMAGE_RAW;
+	if (!strcmp(type, "fax")) return FZ_IMAGE_FAX;
+	if (!strcmp(type, "flate")) return FZ_IMAGE_FLATE;
+	if (!strcmp(type, "lzw")) return FZ_IMAGE_LZW;
+	if (!strcmp(type, "rld")) return FZ_IMAGE_RLD;
+	if (!strcmp(type, "bmp")) return FZ_IMAGE_BMP;
+	if (!strcmp(type, "gif")) return FZ_IMAGE_GIF;
+	if (!strcmp(type, "jbig2")) return FZ_IMAGE_JBIG2;
+	if (!strcmp(type, "jpeg")) return FZ_IMAGE_JPEG;
+	if (!strcmp(type, "jpx")) return FZ_IMAGE_JPX;
+	if (!strcmp(type, "jxr")) return FZ_IMAGE_JXR;
+	if (!strcmp(type, "png")) return FZ_IMAGE_PNG;
+	if (!strcmp(type, "pnm")) return FZ_IMAGE_PNM;
+	if (!strcmp(type, "tiff")) return FZ_IMAGE_TIFF;
+	return FZ_IMAGE_UNKNOWN;
+}
+
 int
 fz_recognize_image_format(fz_context *ctx, unsigned char p[8])
 {
 	if (p[0] == 'P' && p[1] >= '1' && p[1] <= '7')
 		return FZ_IMAGE_PNM;
+	if (p[0] == 'P' && (p[1] == 'F' || p[1] == 'f'))
+		return FZ_IMAGE_PNM;
 	if (p[0] == 0xff && p[1] == 0x4f)
 		return FZ_IMAGE_JPX;
 	if (p[0] == 0x00 && p[1] == 0x00 && p[2] == 0x00 && p[3] == 0x0c &&
@@ -982,6 +1267,11 @@ fz_recognize_image_format(fz_context *ctx, unsigned char p[8])
 		return FZ_IMAGE_GIF;
 	if (p[0] == 'B' && p[1] == 'M')
 		return FZ_IMAGE_BMP;
+	if (p[0] == 'B' && p[1] == 'A')
+		return FZ_IMAGE_BMP;
+	if (p[0] == 0x97 && p[1] == 'J' && p[2] == 'B' && p[3] == '2' &&
+		p[4] == '\r' && p[5] == '\n'  && p[6] == 0x1a && p[7] == '\n')
+		return FZ_IMAGE_JBIG2;
 	return FZ_IMAGE_UNKNOWN;
 }
 
@@ -995,11 +1285,14 @@ fz_new_image_from_buffer(fz_context *ctx, fz_buffer *buffer)
 	unsigned char *buf = buffer->data;
 	fz_image *image = NULL;
 	int type;
+	int bpc;
+	uint8_t orientation = 0;
 
 	if (len < 8)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "unknown image file format");
 
 	type = fz_recognize_image_format(ctx, buf);
+	bpc = 8;
 	switch (type)
 	{
 	case FZ_IMAGE_PNM:
@@ -1009,7 +1302,7 @@ fz_new_image_from_buffer(fz_context *ctx, fz_buffer *buffer)
 		fz_load_jpx_info(ctx, buf, len, &w, &h, &xres, &yres, &cspace);
 		break;
 	case FZ_IMAGE_JPEG:
-		fz_load_jpeg_info(ctx, buf, len, &w, &h, &xres, &yres, &cspace);
+		fz_load_jpeg_info(ctx, buf, len, &w, &h, &xres, &yres, &cspace, &orientation);
 		break;
 	case FZ_IMAGE_PNG:
 		fz_load_png_info(ctx, buf, len, &w, &h, &xres, &yres, &cspace);
@@ -1026,6 +1319,10 @@ fz_new_image_from_buffer(fz_context *ctx, fz_buffer *buffer)
 	case FZ_IMAGE_BMP:
 		fz_load_bmp_info(ctx, buf, len, &w, &h, &xres, &yres, &cspace);
 		break;
+	case FZ_IMAGE_JBIG2:
+		fz_load_jbig2_info(ctx, buf, len, &w, &h, &xres, &yres, &cspace);
+		bpc = 1;
+		break;
 	default:
 		fz_throw(ctx, FZ_ERROR_GENERIC, "unknown image file format");
 	}
@@ -1037,7 +1334,8 @@ fz_new_image_from_buffer(fz_context *ctx, fz_buffer *buffer)
 		bc->params.type = type;
 		if (type == FZ_IMAGE_JPEG)
 			bc->params.u.jpeg.color_transform = -1;
-		image = fz_new_image_from_compressed_buffer(ctx, w, h, 8, cspace, xres, yres, 0, 0, NULL, NULL, bc, NULL);
+		image = fz_new_image_from_compressed_buffer(ctx, w, h, bpc, cspace, xres, yres, 0, 0, NULL, NULL, bc, NULL);
+		image->orientation = orientation;
 	}
 	fz_always(ctx)
 		fz_drop_colorspace(ctx, cspace);
@@ -1086,12 +1384,7 @@ fz_image_resolution(fz_image *image, int *xres, int *yres)
 	/* Scale xres and yres up until we get believable values */
 	if (*xres < SANE_DPI || *yres < SANE_DPI || *xres > INSANE_DPI || *yres > INSANE_DPI)
 	{
-		if (*xres == *yres)
-		{
-			*xres = SANE_DPI;
-			*yres = SANE_DPI;
-		}
-		else if (*xres < *yres)
+		if (*xres < *yres)
 		{
 			*yres = *yres * SANE_DPI / *xres;
 			*xres = SANE_DPI;
@@ -1101,9 +1394,72 @@ fz_image_resolution(fz_image *image, int *xres, int *yres)
 			*xres = *xres * SANE_DPI / *yres;
 			*yres = SANE_DPI;
 		}
+
+		if (*xres == *yres || *xres < SANE_DPI || *yres < SANE_DPI || *xres > INSANE_DPI || *yres > INSANE_DPI)
+		{
+			*xres = SANE_DPI;
+			*yres = SANE_DPI;
+		}
 	}
 }
 
+uint8_t fz_image_orientation(fz_context *ctx, fz_image *image)
+{
+	return image ? image->orientation : 0;
+}
+
+fz_matrix fz_image_orientation_matrix(fz_context *ctx, fz_image *image)
+{
+	fz_matrix m;
+
+	switch (image ? image->orientation : 0)
+	{
+	case 0:
+	case 1: /* 0 degree rotation */
+		m.a =  1; m.b =  0;
+		m.c =  0; m.d =  1;
+		m.e =  0; m.f =  0;
+		break;
+	case 2: /* 90 degree ccw */
+		m.a =  0; m.b = -1;
+		m.c =  1; m.d =  0;
+		m.e =  0; m.f =  1;
+		break;
+	case 3: /* 180 degree ccw */
+		m.a = -1; m.b =  0;
+		m.c =  0; m.d = -1;
+		m.e =  1; m.f =  1;
+		break;
+	case 4: /* 270 degree ccw */
+		m.a =  0; m.b =  1;
+		m.c = -1; m.d =  0;
+		m.e =  1; m.f =  0;
+		break;
+	case 5: /* flip on X */
+		m.a = -1; m.b = 0;
+		m.c =  0; m.d = 1;
+		m.e =  1; m.f = 0;
+		break;
+	case 6: /* flip on X, then rotate ccw by 90 degrees */
+		m.a =  0; m.b =  1;
+		m.c =  1; m.d =  0;
+		m.e =  0; m.f =  0;
+		break;
+	case 7: /* flip on X, then rotate ccw by 180 degrees */
+		m.a =  1; m.b =  0;
+		m.c =  0; m.d = -1;
+		m.e =  0; m.f =  1;
+		break;
+	case 8: /* flip on X, then rotate ccw by 270 degrees */
+		m.a =  0; m.b = -1;
+		m.c = -1; m.d =  0;
+		m.e =  1; m.f =  1;
+		break;
+	}
+
+	return m;
+}
+
 typedef struct fz_display_list_image_s
 {
 	fz_image super;
@@ -1119,6 +1475,8 @@ display_list_image_get_pixmap(fz_context *ctx, fz_image *image_, fz_irect *subar
 	fz_device *dev;
 	fz_pixmap *pix;
 
+	fz_var(dev);
+
 	if (subarea)
 	{
 		/* So, the whole image should be scaled to w * h, but we only want the
@@ -1142,16 +1500,19 @@ display_list_image_get_pixmap(fz_context *ctx, fz_image *image_, fz_irect *subar
 	ctm = fz_pre_scale(image->transform, w, h);
 
 	fz_clear_pixmap(ctx, pix); /* clear to transparent */
-	dev = fz_new_draw_device(ctx, ctm, pix);
 	fz_try(ctx)
 	{
+		dev = fz_new_draw_device(ctx, ctm, pix);
 		fz_run_display_list(ctx, image->list, dev, fz_identity, fz_infinite_rect, NULL);
 		fz_close_device(ctx, dev);
 	}
 	fz_always(ctx)
 		fz_drop_device(ctx, dev);
 	fz_catch(ctx)
+	{
+		fz_drop_pixmap(ctx, pix);
 		fz_rethrow(ctx);
+	}
 
 	/* Never do more subsampling, cos we've already given them the right size */
 	if (l2factor)
