@@ -1,6 +1,29 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
-#include <zlib.h>
+#include "pixmap-imp.h"
+#include "z-imp.h"
 
 #include <limits.h>
 #include <string.h>
@@ -10,7 +33,7 @@ struct info
 	unsigned int width, height, depth, n;
 	enum fz_colorspace_type type;
 	int interlace, indexed;
-	unsigned int size;
+	size_t size;
 	unsigned char *samples;
 	unsigned char palette[256*4];
 	int transparency;
@@ -63,16 +86,6 @@ static const unsigned char png_signature[8] =
 	137, 80, 78, 71, 13, 10, 26, 10
 };
 
-static void *zalloc_png(void *opaque, unsigned int items, unsigned int size)
-{
-	return fz_malloc_array_no_throw(opaque, items, size);
-}
-
-static void zfree_png(void *opaque, void *address)
-{
-	fz_free(opaque, address);
-}
-
 static inline int paeth(int a, int b, int c)
 {
 	/* The definitions of ac and bc are correct, not a typo. */
@@ -188,11 +201,13 @@ png_deinterlace(fz_context *ctx, struct info *info, unsigned int *passw, unsigne
 {
 	unsigned int n = info->n;
 	unsigned int depth = info->depth;
-	unsigned int stride = (info->width * n * depth + 7) / 8;
+	size_t stride = ((size_t)info->width * n * depth + 7) / 8;
 	unsigned char *output;
 	unsigned int p, x, y, k;
 
-	output = fz_malloc_array(ctx, info->height, stride);
+	if (info->height > UINT_MAX / stride)
+		fz_throw(ctx, FZ_ERROR_MEMORY, "image too large");
+	output = Memento_label(fz_malloc(ctx, info->height * stride), "png_deinterlace");
 
 	for (p = 0; p < 7; p++)
 	{
@@ -344,8 +359,10 @@ png_read_trns(fz_context *ctx, struct info *info, const unsigned char *p, unsign
 static void
 png_read_icc(fz_context *ctx, struct info *info, const unsigned char *p, unsigned int size)
 {
+#if FZ_ENABLE_ICC
 	fz_stream *mstm = NULL, *zstm = NULL;
 	fz_colorspace *cs = NULL;
+	fz_buffer *buf = NULL;
 	size_t m = fz_mini(80, size);
 	size_t n = fz_strnlen((const char *)p, m);
 	if (n + 2 > m)
@@ -356,23 +373,26 @@ png_read_icc(fz_context *ctx, struct info *info, const unsigned char *p, unsigne
 
 	fz_var(mstm);
 	fz_var(zstm);
+	fz_var(buf);
 
 	fz_try(ctx)
 	{
 		mstm = fz_open_memory(ctx, p + n + 2, size - n - 2);
 		zstm = fz_open_flated(ctx, mstm, 15);
-		cs = fz_new_icc_colorspace_from_stream(ctx, info->type, zstm);
-		/* drop old one in case we have multiple ICC profiles */
+		buf = fz_read_all(ctx, zstm, 0);
+		cs = fz_new_icc_colorspace(ctx, info->type, 0, NULL, buf);
 		fz_drop_colorspace(ctx, info->cs);
 		info->cs = cs;
 	}
 	fz_always(ctx)
 	{
-		fz_drop_stream(ctx, mstm);
+		fz_drop_buffer(ctx, buf);
 		fz_drop_stream(ctx, zstm);
+		fz_drop_stream(ctx, mstm);
 	}
 	fz_catch(ctx)
-		fz_warn(ctx, "cannot read embedded ICC profile");
+		fz_warn(ctx, "ignoring embedded ICC profile in PNG");
+#endif
 }
 
 static void
@@ -429,7 +449,7 @@ png_read_image(fz_context *ctx, struct info *info, const unsigned char *p, size_
 	/* Read IHDR chunk (must come first) */
 
 	size = getuint(p);
-	if (total < 12 || size > total - 12)
+	if (size > total - 12)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "premature end of data in png image");
 
 	if (!memcmp(p + 4, "IHDR", 4))
@@ -445,7 +465,7 @@ png_read_image(fz_context *ctx, struct info *info, const unsigned char *p, size_
 	{
 		if (!info->interlace)
 		{
-			info->size = info->height * (1 + (info->width * info->n * info->depth + 7) / 8);
+			info->size = info->height * (1 + ((size_t) info->width * info->n * info->depth + 7) / 8);
 		}
 		else
 		{
@@ -453,21 +473,18 @@ png_read_image(fz_context *ctx, struct info *info, const unsigned char *p, size_
 			info->size = passofs[7];
 		}
 
-		info->samples = fz_malloc(ctx, info->size);
+		info->samples = Memento_label(fz_malloc(ctx, info->size), "png_samples");
 
-		stm.zalloc = zalloc_png;
-		stm.zfree = zfree_png;
+		stm.zalloc = fz_zlib_alloc;
+		stm.zfree = fz_zlib_free;
 		stm.opaque = ctx;
 
 		stm.next_out = info->samples;
-		stm.avail_out = info->size;
+		stm.avail_out = (uInt)info->size;
 
 		code = inflateInit(&stm);
 		if (code != Z_OK)
-		{
-			fz_free(ctx, info->samples);
 			fz_throw(ctx, FZ_ERROR_GENERIC, "zlib error: %s", stm.msg);
-		}
 	}
 
 	fz_try(ctx)
@@ -564,8 +581,8 @@ png_expand_palette(fz_context *ctx, struct info *info, fz_pixmap *src)
 	unsigned char *sp = src->samples;
 	unsigned char *dp = dst->samples;
 	unsigned int x, y;
-	int dstride = dst->stride - dst->w * dst->n;
-	int sstride = src->stride - src->w * src->n;
+	size_t dstride = dst->stride - dst->w * (size_t)dst->n;
+	size_t sstride = src->stride - src->w * (size_t)src->n;
 
 	dst->xres = src->xres;
 	dst->yres = src->yres;
@@ -619,7 +636,7 @@ fz_load_png(fz_context *ctx, const unsigned char *p, size_t total)
 {
 	fz_pixmap *image = NULL;
 	struct info png;
-	int stride;
+	size_t stride;
 	int alpha;
 
 	fz_var(image);
@@ -628,7 +645,7 @@ fz_load_png(fz_context *ctx, const unsigned char *p, size_t total)
 	{
 		png_read_image(ctx, &png, p, total, 0);
 
-		stride = (png.width * png.n * png.depth + 7) / 8;
+		stride = ((size_t) png.width * png.n * png.depth + 7) / 8;
 		alpha = (png.n == 2 || png.n == 4 || png.transparency);
 
 		if (png.indexed)
