@@ -1,9 +1,35 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
+#include "color-imp.h"
+#include "pixmap-imp.h"
+
 #include <assert.h>
 #include <limits.h>
 #include <string.h>
 #include <math.h>
+#include <float.h>
 
 fz_pixmap *
 fz_keep_pixmap(fz_context *ctx, fz_pixmap *pix)
@@ -45,6 +71,8 @@ fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal stride for pixmap (n=%d w=%d, stride=%d)", n, w, stride);
 	if (samples == NULL && stride < n*w)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal -ve stride for pixmap without data");
+	if (n > FZ_MAX_COLORS)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal number of colorants");
 
 	pix = fz_malloc_struct(ctx, fz_pixmap);
 	FZ_INIT_STORABLE(pix, 1, fz_drop_pixmap_imp);
@@ -72,17 +100,18 @@ fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h
 	}
 
 	pix->samples = samples;
-	if (!samples)
+	if (!samples && pix->h > 0 && pix->w > 0)
 	{
 		fz_try(ctx)
 		{
-			if (pix->stride - 1 > INT_MAX / pix->n)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "overly wide image");
-			pix->samples = fz_malloc_array(ctx, pix->h, pix->stride);
+			if ((size_t)pix->stride > SIZE_MAX / (size_t)pix->h)
+				fz_throw(ctx, FZ_ERROR_GENERIC, "Overly large image");
+			pix->samples = Memento_label(fz_malloc(ctx, pix->h * pix->stride), "pixmap_data");
 		}
 		fz_catch(ctx)
 		{
-			fz_drop_colorspace(ctx, colorspace);
+			fz_drop_separations(ctx, pix->seps);
+			fz_drop_colorspace(ctx, pix->colorspace);
 			fz_free(ctx, pix);
 			fz_rethrow(ctx);
 		}
@@ -97,8 +126,12 @@ fz_new_pixmap(fz_context *ctx, fz_colorspace *colorspace, int w, int h, fz_separ
 {
 	int stride;
 	int s = fz_count_active_separations(ctx, seps);
+	int n;
 	if (!colorspace && s == 0) alpha = 1;
-	stride = (fz_colorspace_n(ctx, colorspace) + s + alpha) * w;
+	n = fz_colorspace_n(ctx, colorspace) + s + alpha;
+	if (w > INT_MAX / n)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Overly wide image");
+	stride = n * w;
 	return fz_new_pixmap_with_data(ctx, colorspace, w, h, seps, alpha, stride, NULL);
 }
 
@@ -106,7 +139,7 @@ fz_pixmap *
 fz_new_pixmap_with_bbox(fz_context *ctx, fz_colorspace *colorspace, fz_irect bbox, fz_separations *seps, int alpha)
 {
 	fz_pixmap *pixmap;
-	pixmap = fz_new_pixmap(ctx, colorspace, bbox.x1 - bbox.x0, bbox.y1 - bbox.y0, seps, alpha);
+	pixmap = fz_new_pixmap(ctx, colorspace, fz_irect_width(bbox), fz_irect_height(bbox), seps, alpha);
 	pixmap->x = bbox.x0;
 	pixmap->y = bbox.y0;
 	return pixmap;
@@ -115,13 +148,13 @@ fz_new_pixmap_with_bbox(fz_context *ctx, fz_colorspace *colorspace, fz_irect bbo
 fz_pixmap *
 fz_new_pixmap_with_bbox_and_data(fz_context *ctx, fz_colorspace *colorspace, fz_irect bbox, fz_separations *seps, int alpha, unsigned char *samples)
 {
-	int w = bbox.x1 - bbox.x0;
+	int w = fz_irect_width(bbox);
 	int stride;
 	int s = fz_count_active_separations(ctx, seps);
 	fz_pixmap *pixmap;
 	if (!colorspace && s == 0) alpha = 1;
 	stride = (fz_colorspace_n(ctx, colorspace) + s + alpha) * w;
-	pixmap = fz_new_pixmap_with_data(ctx, colorspace, w, bbox.y1 - bbox.y0, seps, alpha, stride, samples);
+	pixmap = fz_new_pixmap_with_data(ctx, colorspace, w, fz_irect_height(bbox), seps, alpha, stride, samples);
 	pixmap->x = bbox.x0;
 	pixmap->y = bbox.y0;
 	return pixmap;
@@ -151,8 +184,8 @@ fz_pixmap *fz_new_pixmap_from_pixmap(fz_context *ctx, fz_pixmap *pixmap, const f
 	subpix->storable.refs = 1;
 	subpix->x = rect->x0;
 	subpix->y = rect->y0;
-	subpix->w = rect->x1 - rect->x0;
-	subpix->h = rect->y1 - rect->y0;
+	subpix->w = fz_irect_width(*rect);
+	subpix->h = fz_irect_height(*rect);
 	subpix->samples += (rect->x0 - pixmap->x) + (rect->y0 - pixmap->y) * pixmap->stride;
 	subpix->underlying = fz_keep_pixmap(ctx, pixmap);
 	subpix->colorspace = fz_keep_colorspace(ctx, pixmap->colorspace);
@@ -162,6 +195,13 @@ fz_pixmap *fz_new_pixmap_from_pixmap(fz_context *ctx, fz_pixmap *pixmap, const f
 	return subpix;
 }
 
+fz_pixmap *fz_clone_pixmap(fz_context *ctx, const fz_pixmap *old)
+{
+	fz_pixmap *pix = fz_new_pixmap_with_bbox(ctx, old->colorspace, fz_make_irect(old->x, old->y, old->w, old->h), old->seps, old->alpha);
+	memcpy(pix->samples, old->samples, pix->stride * pix->h);
+	return pix;
+}
+
 fz_irect
 fz_pixmap_bbox(fz_context *ctx, const fz_pixmap *pix)
 {
@@ -185,7 +225,7 @@ fz_pixmap_bbox_no_ctx(const fz_pixmap *pix)
 }
 
 fz_colorspace *
-fz_pixmap_colorspace(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_colorspace(fz_context *ctx, const fz_pixmap *pix)
 {
 	if (!pix)
 		return NULL;
@@ -193,61 +233,61 @@ fz_pixmap_colorspace(fz_context *ctx, fz_pixmap *pix)
 }
 
 int
-fz_pixmap_x(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_x(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->x;
 }
 
 int
-fz_pixmap_y(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_y(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->y;
 }
 
 int
-fz_pixmap_width(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_width(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->w;
 }
 
 int
-fz_pixmap_height(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_height(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->h;
 }
 
 int
-fz_pixmap_components(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_components(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->n;
 }
 
 int
-fz_pixmap_colorants(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_colorants(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->n - pix->alpha - pix->s;
 }
 
 int
-fz_pixmap_spots(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_spots(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->s;
 }
 
 int
-fz_pixmap_alpha(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_alpha(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->alpha;
 }
 
 int
-fz_pixmap_stride(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_stride(fz_context *ctx, const fz_pixmap *pix)
 {
 	return pix->stride;
 }
 
 unsigned char *
-fz_pixmap_samples(fz_context *ctx, fz_pixmap *pix)
+fz_pixmap_samples(fz_context *ctx, const fz_pixmap *pix)
 {
 	if (!pix)
 		return NULL;
@@ -480,7 +520,7 @@ clear_cmyk_bitmap(unsigned char *samples, int w, int h, int spots, int stride, i
 void
 fz_clear_pixmap(fz_context *ctx, fz_pixmap *pix)
 {
-	int stride = pix->w * pix->n;
+	ptrdiff_t stride = pix->w * (ptrdiff_t)pix->n;
 	int h = pix->h;
 	unsigned char *s = pix->samples;
 	if (stride == pix->stride)
@@ -492,27 +532,55 @@ fz_clear_pixmap(fz_context *ctx, fz_pixmap *pix)
 	{
 		while (h--)
 		{
-			memset(s, 0, (unsigned int)stride);
+			memset(s, 0, stride);
+			s += pix->stride;
+		}
+	}
+	else if (pix->s == 0)
+	{
+		while (h--)
+		{
+			memset(s, 0xff, stride);
 			s += pix->stride;
 		}
 	}
 	else
 	{
+		/* Horrible, slow case: additive with spots */
+		size_t w = stride/pix->n;
+		int spots = pix->s;
+		int colorants = pix->n - spots; /* We know there is no alpha */
 		while (h--)
 		{
-			memset(s, 0xff, (unsigned int)stride);
-			s += pix->stride;
+			size_t w2 = w;
+			while (w2--)
+			{
+				int i = colorants;
+				do
+				{
+					*s++ = 0xff;
+					i--;
+				}
+				while (i != 0);
+
+				i = spots;
+				do
+				{
+					*s++ = 0;
+					i--;
+				}
+				while (i != 0);
+			}
 		}
 	}
 }
 
-/* This function is horrible, and should be removed from the
- * API and replaced with a less magic one. */
 void
 fz_clear_pixmap_with_value(fz_context *ctx, fz_pixmap *pix, int value)
 {
 	unsigned char *s;
-	int w, h, n, stride, len;
+	int w, h, n;
+	ptrdiff_t stride, len;
 	int alpha = pix->alpha;
 
 	w = pix->w;
@@ -529,7 +597,7 @@ fz_clear_pixmap_with_value(fz_context *ctx, fz_pixmap *pix, int value)
 
 	n = pix->n;
 	stride = pix->stride;
-	len = w * n;
+	len = (ptrdiff_t)w * n;
 
 	s = pix->samples;
 	if (value == 255 || !alpha)
@@ -541,7 +609,7 @@ fz_clear_pixmap_with_value(fz_context *ctx, fz_pixmap *pix, int value)
 		}
 		while (h--)
 		{
-			memset(s, value, (unsigned int)len);
+			memset(s, value, len);
 			s += stride;
 		}
 	}
@@ -564,7 +632,7 @@ fz_clear_pixmap_with_value(fz_context *ctx, fz_pixmap *pix, int value)
 }
 
 void
-fz_fill_pixmap_with_color(fz_context *ctx, fz_pixmap *pix, fz_colorspace *colorspace, float *color, const fz_color_params *color_params)
+fz_fill_pixmap_with_color(fz_context *ctx, fz_pixmap *pix, fz_colorspace *colorspace, float *color, fz_color_params color_params)
 {
 	float colorfv[FZ_MAX_COLORS];
 	unsigned char colorbv[FZ_MAX_COLORS];
@@ -573,7 +641,7 @@ fz_fill_pixmap_with_color(fz_context *ctx, fz_pixmap *pix, fz_colorspace *colors
 	n = fz_colorspace_n(ctx, pix->colorspace);
 	a = pix->alpha;
 	s = pix->s;
-	fz_convert_color(ctx, color_params, NULL, pix->colorspace, colorfv, colorspace, color);
+	fz_convert_color(ctx, colorspace, color, pix->colorspace, colorfv, NULL, color_params);
 	for (i = 0; i < n; ++i)
 		colorbv[i] = colorfv[i] * 255;
 
@@ -599,19 +667,20 @@ fz_copy_pixmap_rect(fz_context *ctx, fz_pixmap *dest, fz_pixmap *src, fz_irect b
 {
 	unsigned char *srcp;
 	unsigned char *destp;
-	int y, w, destspan, srcspan;
+	unsigned int y, w;
+	size_t destspan, srcspan;
 
 	b = fz_intersect_irect(b, fz_pixmap_bbox(ctx, dest));
 	b = fz_intersect_irect(b, fz_pixmap_bbox(ctx, src));
-	w = b.x1 - b.x0;
-	y = b.y1 - b.y0;
-	if (w <= 0 || y <= 0)
+	if (fz_is_empty_irect(b))
 		return;
+	w = (unsigned int)(b.x1 - b.x0);
+	y = (unsigned int)(b.y1 - b.y0);
 
 	srcspan = src->stride;
-	srcp = src->samples + (unsigned int)(srcspan * (b.y0 - src->y) + src->n * (b.x0 - src->x));
+	srcp = src->samples + srcspan * (b.y0 - src->y) + (b.x0 - src->x) * (size_t)src->n;
 	destspan = dest->stride;
-	destp = dest->samples + (unsigned int)(destspan * (b.y0 - dest->y) + dest->n * (b.x0 - dest->x));
+	destp = dest->samples + destspan * (b.y0 - dest->y) + (b.x0 - dest->x) * (size_t)dest->n;
 
 	if (src->n == dest->n)
 	{
@@ -626,16 +695,13 @@ fz_copy_pixmap_rect(fz_context *ctx, fz_pixmap *dest, fz_pixmap *src, fz_irect b
 	}
 	else
 	{
-		fz_pixmap_converter *pc = fz_lookup_pixmap_converter(ctx, dest->colorspace, src->colorspace);
 		fz_pixmap fake_src = *src;
-
 		fake_src.x = b.x0;
 		fake_src.y = b.y0;
 		fake_src.w = w;
 		fake_src.h = y;
 		fake_src.samples = srcp;
-
-		pc(ctx, dest, &fake_src, NULL, default_cs, fz_default_color_params(ctx), 0);
+		fz_convert_pixmap_samples(ctx, &fake_src, dest, NULL, default_cs, fz_default_color_params, 0);
 	}
 }
 
@@ -643,7 +709,8 @@ void
 fz_clear_pixmap_rect_with_value(fz_context *ctx, fz_pixmap *dest, int value, fz_irect b)
 {
 	unsigned char *destp;
-	int x, y, w, k, destspan;
+	int x, y, w, k;
+	size_t destspan;
 
 	b = fz_intersect_irect(b, fz_pixmap_bbox(ctx, dest));
 	w = b.x1 - b.x0;
@@ -652,7 +719,7 @@ fz_clear_pixmap_rect_with_value(fz_context *ctx, fz_pixmap *dest, int value, fz_
 		return;
 
 	destspan = dest->stride;
-	destp = dest->samples + (unsigned int)(destspan * (b.y0 - dest->y) + dest->n * (b.x0 - dest->x));
+	destp = dest->samples + destspan * (b.y0 - dest->y) + (b.x0 - dest->x) * (size_t)dest->n;
 
 	/* CMYK needs special handling (and potentially any other subtractive colorspaces) */
 	if (fz_colorspace_n(ctx, dest->colorspace) == 4)
@@ -679,7 +746,7 @@ fz_clear_pixmap_rect_with_value(fz_context *ctx, fz_pixmap *dest, int value, fz_
 	{
 		do
 		{
-			memset(destp, 255, (unsigned int)(w * dest->n));
+			memset(destp, 255, w * (size_t)dest->n);
 			destp += destspan;
 		}
 		while (--y);
@@ -707,7 +774,7 @@ fz_premultiply_pixmap(fz_context *ctx, fz_pixmap *pix)
 	unsigned char *s = pix->samples;
 	unsigned char a;
 	int k, x, y;
-	int stride = pix->stride - pix->w * pix->n;
+	size_t stride = pix->stride - pix->w * (size_t)pix->n;
 
 	if (!pix->alpha)
 		return;
@@ -753,21 +820,32 @@ fz_alpha_from_gray(fz_context *ctx, fz_pixmap *gray)
 }
 
 void
-fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int r, int g, int b)
+fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int black, int white)
 {
 	unsigned char *s = pix->samples;
 	int n = pix->n;
 	int x, y, save;
+	int rb = (black>>16)&255;
+	int gb = (black>>8)&255;
+	int bb = (black)&255;
+	int rw = (white>>16)&255;
+	int gw = (white>>8)&255;
+	int bw = (white)&255;
+	int rm = (rw - rb);
+	int gm = (gw - gb);
+	int bm = (bw - bb);
 
 	switch (fz_colorspace_type(ctx, pix->colorspace))
 	{
 	case FZ_COLORSPACE_GRAY:
-		g = (r + g + b) / 3;
+		gw = (rw + gw + bw) / 3;
+		gb = (rb + gb + bb) / 3;
+		gm = gw - gb;
 		for (y = 0; y < pix->h; y++)
 		{
 			for (x = 0; x < pix->w; x++)
 			{
-				*s = fz_mul255(*s, g);
+				*s = gb + fz_mul255(*s, gm);
 				s += n;
 			}
 			s += pix->stride - pix->w * n;
@@ -775,18 +853,17 @@ fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int r, int g, int b)
 		break;
 
 	case FZ_COLORSPACE_BGR:
-		save = r;
-		r = b;
-		b = save;
+		save = rm; rm = bm; bm = save;
+		save = rb; rb = bb; bb = save;
 		/* fall through */
 	case FZ_COLORSPACE_RGB:
 		for (y = 0; y < pix->h; y++)
 		{
 			for (x = 0; x < pix->w; x++)
 			{
-				s[0] = fz_mul255(s[0], r);
-				s[1] = fz_mul255(s[1], g);
-				s[2] = fz_mul255(s[2], b);
+				s[0] = rb + fz_mul255(s[0], rm);
+				s[1] = gb + fz_mul255(s[1], gm);
+				s[2] = bb + fz_mul255(s[2], bm);
 				s += n;
 			}
 			s += pix->stride - pix->w * n;
@@ -799,8 +876,232 @@ fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int r, int g, int b)
 	}
 }
 
+/* Invert luminance in RGB/BGR pixmap, but keep the colors as is. */
+static inline void invert_luminance(int type, unsigned char *s)
+{
+	int r, g, b, y;
+
+	/* Convert to YUV */
+	if (type == FZ_COLORSPACE_RGB)
+	{
+		r = s[0];
+		g = s[1];
+		b = s[2];
+	}
+	else
+	{
+		r = s[2];
+		g = s[1];
+		b = s[0];
+	}
+
+	y = (39336 * r + 76884 * g + 14900 * b + 32768)>>16;
+	y = 259-y;
+	r += y;
+	g += y;
+	b += y;
+
+	if (type == FZ_COLORSPACE_RGB)
+	{
+		s[0] = r > 255 ? 255 : r < 0 ? 0 : r;
+		s[1] = g > 255 ? 255 : g < 0 ? 0 : g;
+		s[2] = b > 255 ? 255 : b < 0 ? 0 : b;
+	}
+	else
+	{
+		s[2] = r > 255 ? 255 : r < 0 ? 0 : r;
+		s[1] = g > 255 ? 255 : g < 0 ? 0 : g;
+		s[0] = b > 255 ? 255 : b < 0 ? 0 : b;
+	}
+}
+
+void
+fz_invert_pixmap_luminance(fz_context *ctx, fz_pixmap *pix)
+{
+	unsigned char *s = pix->samples;
+	int x, y, n = pix->n;
+	int type = pix->colorspace ? pix->colorspace->type : FZ_COLORSPACE_NONE;
+
+	if (type == FZ_COLORSPACE_GRAY)
+	{
+		fz_invert_pixmap(ctx, pix);
+	}
+	else if (type == FZ_COLORSPACE_RGB || type == FZ_COLORSPACE_BGR)
+	{
+		for (y = 0; y < pix->h; y++)
+		{
+			for (x = 0; x < pix->w; x++)
+			{
+				invert_luminance(type, s);
+				s += n;
+			}
+			s += pix->stride - pix->w * n;
+		}
+	}
+	else
+	{
+		fz_throw(ctx, FZ_ERROR_GENERIC, "can only invert luminance of Gray and RGB pixmaps");
+	}
+}
+
 void
 fz_invert_pixmap(fz_context *ctx, fz_pixmap *pix)
+{
+	fz_irect rect = { pix->x, pix->y, pix->x + pix->w, pix->y + pix->h };
+	fz_invert_pixmap_rect(ctx, pix, rect);
+}
+
+void
+fz_invert_pixmap_alpha(fz_context *ctx, fz_pixmap *pix)
+{
+	unsigned char *s = pix->samples;
+	int x, y;
+	int n1 = pix->n - pix->alpha;
+	int n = pix->n;
+
+	if (!pix->alpha)
+		return;
+
+	for (y = 0; y < pix->h; y++)
+	{
+		s += n1;
+		for (x = 0; x < pix->w; x++)
+		{
+			*s = 255 - *s;
+			s += n;
+		}
+		s += pix->stride - pix->w * n;
+	}
+}
+
+void fz_invert_pixmap_rect(fz_context *ctx, fz_pixmap *pix, fz_irect rect)
+{
+	int x0 = fz_clampi(rect.x0 - pix->x, 0, pix->w);
+	int x1 = fz_clampi(rect.x1 - pix->x, 0, pix->w);
+	int y0 = fz_clampi(rect.y0 - pix->y, 0, pix->h);
+	int y1 = fz_clampi(rect.y1 - pix->y, 0, pix->h);
+
+	int k, x, y;
+	int n = pix->n;
+	int s = pix->s;
+	int cmyk = (pix->colorspace && pix->colorspace->type == FZ_COLORSPACE_CMYK);
+
+	if (cmyk)
+	{
+		/* For cmyk, we're storing: (a.c, a.m, a.y, a.k, a)
+		 * So, a.r = a - a.c - a.k
+		 *     a.g = a - a.m - a.k
+		 *     a.b = a - a.y - a.k
+		 * Invert that:
+		 *     a.R = a.c + a.k
+		 *     a.G = a.m + a.k
+		 *     a.B = a.y + a.k
+		 * Convert that back to cmy
+		 *     a.C = a - a.c - a.k;
+		 *     a.M = a - a.m - a.k;
+		 *     a.Y = a - a.y - a.k;
+		 * Extract K:
+		 *     a.K' = min(a.C, a.M, a.Y)
+		 *          = a - a.k - max(a.c, a.m, a.y)
+		 *     a.C' = a.C - a.K' = a - a.c - a.k - (a - a.k - max(a.c, a.m, a.y)) = max(a.c, a.m, a.y) - a.c
+		 *     a.M' = a.M - a.K' = a - a.m - a.k - (a - a.k - max(a.c, a.m, a.y)) = max(a.c, a.m, a.y) - a.m
+		 *     a.Y' = a.Y - a.K' = a - a.y - a.k - (a - a.k - max(a.c, a.m, a.y)) = max(a.c, a.m, a.y) - a.y
+		 * */
+		if (pix->alpha)
+		{
+			int n1 = pix->n - pix->alpha - s;
+			for (y = y0; y < y1; y++)
+			{
+				unsigned char *d = pix->samples + ((y * (size_t)pix->stride) + (x0 * (size_t)pix->n));
+				for (x = x0; x < x1; x++)
+				{
+					int ac = d[0];
+					int am = d[1];
+					int ay = d[2];
+					int ak = d[3];
+					int a = d[n1];
+					int mx = fz_maxi(fz_maxi(ac, am), ay);
+					d[0] = mx-ac;
+					d[1] = mx-am;
+					d[2] = mx-ay;
+					ak = a - ak - mx;
+					if (ak < 0)
+						ak = 0;
+					d[3] = ak;
+					d += n;
+				}
+			}
+		}
+		else
+		{
+			for (y = y0; y < y1; y++)
+			{
+				unsigned char *d = pix->samples + ((y * (size_t)pix->stride) + (x0 * (size_t)pix->n));
+				for (x = x0; x < x1; x++)
+				{
+					int c = d[0];
+					int m = d[1];
+					int y = d[2];
+					int k = d[3];
+					int mx = fz_maxi(fz_maxi(c, m), y);
+					d[0] = mx-c;
+					d[1] = mx-m;
+					d[2] = mx-y;
+					k = 255 - k - mx;
+					if (k < 0)
+						k = 0;
+					d[3] = k;
+					d += n;
+				}
+			}
+		}
+	}
+	else if (pix->alpha)
+	{
+		int n1 = pix->n - pix->alpha - s;
+		for (y = y0; y < y1; y++)
+		{
+			unsigned char *d = pix->samples + ((y * (size_t)pix->stride) + (x0 * (size_t)pix->n));
+			for (x = x0; x < x1; x++)
+			{
+				int a = d[n1];
+				for (k = 0; k < n1; k++)
+					d[k] = a - d[k];
+				d += n;
+			}
+		}
+	}
+	else if (s)
+	{
+		int n1 = pix->n - s;
+		for (y = y0; y < y1; y++)
+		{
+			unsigned char *d = pix->samples + ((y * (size_t)pix->stride) + (x0 * (size_t)pix->n));
+			for (x = x0; x < x1; x++)
+			{
+				for (k = 0; k < n1; k++)
+					d[k] = 255 - d[k];
+				d += n;
+			}
+		}
+	}
+	else
+	{
+		for (y = y0; y < y1; y++)
+		{
+			unsigned char *d = pix->samples + ((y * (size_t)pix->stride) + (x0 * (size_t)pix->n));
+			for (x = x0; x < x1; x++)
+			{
+				for (k = 0; k < n; k++)
+					d[k] = 255 - d[k];
+				d += n;
+			}
+		}
+	}
+}
+
+void
+fz_invert_pixmap_raw(fz_context *ctx, fz_pixmap *pix)
 {
 	unsigned char *s = pix->samples;
 	int k, x, y;
@@ -819,28 +1120,6 @@ fz_invert_pixmap(fz_context *ctx, fz_pixmap *pix)
 	}
 }
 
-void fz_invert_pixmap_rect(fz_context *ctx, fz_pixmap *image, fz_irect rect)
-{
-	unsigned char *p;
-	int x, y, n;
-
-	int x0 = fz_clampi(rect.x0 - image->x, 0, image->w);
-	int x1 = fz_clampi(rect.x1 - image->x, 0, image->w);
-	int y0 = fz_clampi(rect.y0 - image->y, 0, image->h);
-	int y1 = fz_clampi(rect.y1 - image->y, 0, image->h);
-
-	for (y = y0; y < y1; y++)
-	{
-		p = image->samples + (unsigned int)((y * image->stride) + (x0 * image->n));
-		for (x = x0; x < x1; x++)
-		{
-			for (n = image->n; n > 1; n--, p++)
-				*p = 255 - *p;
-			p++;
-		}
-	}
-}
-
 void
 fz_gamma_pixmap(fz_context *ctx, fz_pixmap *pix, float gamma)
 {
@@ -870,20 +1149,17 @@ fz_pixmap_size(fz_context *ctx, fz_pixmap * pix)
 {
 	if (pix == NULL)
 		return 0;
-	return sizeof(*pix) + pix->n * pix->w * pix->h;
+	return sizeof(*pix) + (size_t)pix->n * pix->w * pix->h;
 }
 
 fz_pixmap *
-fz_convert_pixmap(fz_context *ctx, fz_pixmap *pix, fz_colorspace *ds, fz_colorspace *prf, fz_default_colorspaces *default_cs, const fz_color_params *color_params, int keep_alpha)
+fz_convert_pixmap(fz_context *ctx, const fz_pixmap *pix, fz_colorspace *ds, fz_colorspace *prf, fz_default_colorspaces *default_cs, fz_color_params color_params, int keep_alpha)
 {
 	fz_pixmap *cvt;
 
 	if (!ds && !keep_alpha)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot both throw away and keep alpha");
 
-	if (color_params == NULL)
-		color_params = fz_default_color_params(ctx);
-
 	cvt = fz_new_pixmap(ctx, ds, pix->w, pix->h, pix->seps, keep_alpha && pix->alpha);
 
 	cvt->xres = pix->xres;
@@ -897,8 +1173,7 @@ fz_convert_pixmap(fz_context *ctx, fz_pixmap *pix, fz_colorspace *ds, fz_colorsp
 
 	fz_try(ctx)
 	{
-		fz_pixmap_converter *pc = fz_lookup_pixmap_converter(ctx, ds, pix->colorspace);
-		pc(ctx, cvt, pix, prf, default_cs, color_params, 1);
+		fz_convert_pixmap_samples(ctx, pix, cvt, prf, default_cs, color_params, 1);
 	}
 	fz_catch(ctx)
 	{
@@ -954,6 +1229,242 @@ fz_new_pixmap_from_1bpp_data(fz_context *ctx, int x, int y, int w, int h, unsign
 	return pixmap;
 }
 
+static float
+calc_percentile(int *hist, float thr, float scale, float minval, float maxval)
+{
+	float prct;
+	int k = 0, count = 0;
+
+	while (count < thr)
+		count += hist[k++];
+
+	if (k <= 0)
+		prct = k;
+	else
+	{
+		float c0 = count - thr;
+		float c1 = thr - (count - hist[k - 1]);
+		prct = (c1 * k + c0 * (k - 1)) / (c0 + c1);
+	}
+
+	prct /= scale;
+	prct += minval;
+	return fz_clamp(prct, minval, maxval);
+}
+
+static void
+calc_percentiles(fz_context *ctx, int nsamples, float *samples, float *minprct, float *maxprct)
+{
+	float minval, maxval, scale;
+	int *hist, size, k;
+
+	minval = maxval = samples[0];
+	for (k = 1; k < nsamples; k++)
+	{
+		minval = fz_min(minval, samples[k]);
+		maxval = fz_max(maxval, samples[k]);
+	}
+
+	if (minval - maxval == 0)
+	{
+		*minprct = *maxprct = minval;
+		return;
+	}
+
+	size = fz_mini(65535, nsamples);
+	scale = (size - 1) / (maxval - minval);
+
+	hist = fz_calloc(ctx, size, sizeof(int));
+
+	fz_try(ctx)
+	{
+		for (k = 0; k < nsamples; k++)
+			hist[(uint16_t) (scale * (samples[k] - minval))]++;
+
+		*minprct = calc_percentile(hist, 0.01f * nsamples, scale, minval, maxval);
+		*maxprct = calc_percentile(hist, 0.99f * nsamples, scale, minval, maxval);
+	}
+	fz_always(ctx)
+		fz_free(ctx, hist);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
+/* Tone mapping according to "Consistent Tone Reproduction" by Min H. Kim and Jan Kautz. */
+fz_pixmap *
+fz_new_pixmap_from_float_data(fz_context *ctx, fz_colorspace *cs, int w, int h, float *samples)
+{
+	fz_pixmap *pixmap = NULL;
+	unsigned char *dp;
+	float *sample;
+	float *lsamples = NULL;
+	float minsample, maxsample, mu;
+	float k1, d0, sigma, sigmasq2;
+	float minprct, maxprct, range;
+	int y, k, n = fz_colorspace_n(ctx, cs);
+	int nsamples = w * h * n;
+#define KIMKAUTZC1 (3.0f)
+#define KIMKAUTZC2 (0.5f)
+#define MAXLD (logf(300.0f))
+#define MINLD (logf(0.3f))
+
+	fz_var(pixmap);
+	fz_var(lsamples);
+
+	fz_try(ctx)
+	{
+		lsamples = fz_malloc(ctx, nsamples * sizeof(float));
+
+		mu = 0;
+		minsample = FLT_MAX;
+		maxsample = -FLT_MAX;
+
+		for (k = 0; k < nsamples; k++)
+		{
+			lsamples[k] = logf(samples[k] == 0 ? FLT_MIN : samples[k]);
+			mu += lsamples[k];
+			minsample = fz_min(minsample, lsamples[k]);
+			maxsample = fz_max(maxsample, lsamples[k]);
+		}
+
+		mu /= nsamples;
+		d0 = maxsample - minsample;
+		k1 = (MAXLD - MINLD) / d0;
+		sigma = d0 / KIMKAUTZC1;
+		sigmasq2 = sigma * sigma * 2;
+
+		for (k = 0; k < nsamples; k++)
+		{
+			float samplemu = samples[k] - mu;
+			float samplemu2 = samplemu * samplemu;
+			float fw = expf(-samplemu2 / sigmasq2);
+			float k2 = (1 - k1) * fw + k1;
+			samples[k] = expf(KIMKAUTZC2 * k2 * (lsamples[k] - mu) + mu);
+		}
+
+		calc_percentiles(ctx, nsamples, samples, &minprct, &maxprct);
+		range = maxprct - minprct;
+
+		pixmap = fz_new_pixmap(ctx, cs, w, h, NULL, 0);
+
+		dp = pixmap->samples + pixmap->stride * (h - 1);
+		sample = samples;
+
+		for (y = 0; y < h; y++)
+		{
+			unsigned char *dpp = dp;
+
+			for (k = 0; k < w * n; k++)
+				*dpp++ = 255.0f * (fz_clamp(*sample++, minprct, maxprct) - minprct) / range;
+
+			dp -= pixmap->stride;
+		}
+	}
+	fz_always(ctx)
+		fz_free(ctx, lsamples);
+	fz_catch(ctx)
+	{
+		fz_drop_pixmap(ctx, pixmap);
+		fz_rethrow(ctx);
+	}
+
+	return pixmap;
+}
+
+fz_pixmap *
+fz_new_pixmap_from_alpha_channel(fz_context *ctx, fz_pixmap *src)
+{
+	fz_pixmap *dst;
+	int w, h, n, x;
+	unsigned char *sp, *dp;
+
+	if (!src->alpha)
+		return NULL;
+
+	dst = fz_new_pixmap_with_bbox(ctx, NULL, fz_pixmap_bbox(ctx, src), NULL, 1);
+	w = src->w;
+	h = src->h;
+	n = src->n;
+	sp = src->samples + n - 1;
+	dp = dst->samples;
+
+	while (h--)
+	{
+		unsigned char *s = sp;
+		unsigned char *d = dp;
+		for (x = 0; x < w; ++x)
+		{
+			*d++ = *s;
+			s += n;
+		}
+		sp += src->stride;
+		dp += dst->stride;
+	}
+
+	return dst;
+}
+
+fz_pixmap *
+fz_new_pixmap_from_color_and_mask(fz_context *ctx, fz_pixmap *color, fz_pixmap *mask)
+{
+	fz_pixmap *dst;
+	int w = color->w;
+	int h = color->h;
+	int n = color->n;
+	int x, y, k;
+
+	if (color->alpha)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "color pixmap must not have an alpha channel");
+	if (mask->n != 1)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "mask pixmap must have exactly one channel");
+	if (mask->w != color->w || mask->h != color->h)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "color and mask pixmaps must be the same size");
+
+	dst = fz_new_pixmap_with_bbox(ctx, color->colorspace, fz_pixmap_bbox(ctx, color), NULL, 1);
+
+	for (y = 0; y < h; ++y)
+	{
+		unsigned char *cs = &color->samples[y * color->stride];
+		unsigned char *ms = &mask->samples[y * mask->stride];
+		unsigned char *ds = &dst->samples[y * dst->stride];
+		for (x = 0; x < w; ++x)
+		{
+			unsigned char a = *ms++;
+			for (k = 0; k < n; ++k)
+				*ds++ = fz_mul255(*cs++, a);
+			*ds++ = a;
+		}
+	}
+
+	return dst;
+}
+
+int
+fz_is_pixmap_monochrome(fz_context *ctx, fz_pixmap *pixmap)
+{
+	int n = pixmap->n;
+	int w = pixmap->w;
+	int h = pixmap->h;
+	unsigned char *s = pixmap->samples;
+	int x;
+
+	if (n != 1)
+		return 0;
+
+	while (h--)
+	{
+		for (x = 0; x < w; ++x)
+		{
+			unsigned char v = s[x];
+			if (v != 0 && v != 255)
+				return 0;
+		}
+		s += pixmap->stride;
+	}
+
+	return 1;
+}
+
 #ifdef ARCH_ARM
 static void
 fz_subsample_pixmap_ARM(unsigned char *ptr, int w, int h, int f, int factor,
@@ -980,7 +1491,7 @@ fz_subsample_pixmap_ARM(unsigned char *ptr, int w, int h, int f, int factor,
 	"ldr	r6, [r13,#4*12]		@ r6 = fwd			\n"
 	"ldr	r7, [r13,#4*13]		@ r7 = back			\n"
 	"subs	r2, r2, r3		@ r2 = h -= f			\n"
-	"blt	11f			@ Skip if less than a full row	\n"
+	"blt	12f			@ Skip if less than a full row	\n"
 	"1:				@ for (y = h; y > 0; y--) {	\n"
 	"ldr	r1, [r13]		@ r1 = w			\n"
 	"subs	r1, r1, r3		@ r1 = w -= f			\n"
@@ -1051,6 +1562,7 @@ fz_subsample_pixmap_ARM(unsigned char *ptr, int w, int h, int f, int factor,
 	"subs	r2, r2, r3		@ h -= f			\n"
 	"add	r0, r0, r14		@ s += fwd3			\n"
 	"bge	1b			@ }				\n"
+	"12:								\n"
 	"adds	r2, r2, r3		@ h += f			\n"
 	"beq	21f			@ if no stray row, end		\n"
 	"@ So doing one last (partial) row				\n"
@@ -1062,7 +1574,7 @@ fz_subsample_pixmap_ARM(unsigned char *ptr, int w, int h, int f, int factor,
 	"@ r4 = factor							\n"
 	"@ r5 = n							\n"
 	"@ r6 = fwd							\n"
-	"12:				@ for (y = h; y > 0; y--) {	\n"
+	"				@ for (y = h; y > 0; y--) {	\n"
 	"ldr	r1, [r13]		@ r1 = w			\n"
 	"ldr	r7, [r13,#4*21]		@ r7 = back5			\n"
 	"ldr	r8, [r13,#4*14]		@ r8 = back2			\n"
@@ -1139,29 +1651,41 @@ fz_subsample_pixmap_ARM(unsigned char *ptr, int w, int h, int f, int factor,
 void
 fz_subsample_pixmap(fz_context *ctx, fz_pixmap *tile, int factor)
 {
-	int dst_w, dst_h, w, h, fwd, fwd2, fwd3, back, back2, n, f;
-	unsigned char *s, *d;
-#ifndef ARCH_ARM
-	int x, y, xx, yy, nn;
-#endif
+	int f;
 
 	if (!tile)
 		return;
 
 	assert(tile->stride >= tile->w * tile->n);
 
-	s = d = tile->samples;
+	fz_subsample_pixblock(tile->samples, tile->w, tile->h, tile->n, factor, tile->stride);
+
 	f = 1<<factor;
-	w = tile->w;
-	h = tile->h;
-	n = tile->n;
-	dst_w = (w + f-1)>>factor;
-	dst_h = (h + f-1)>>factor;
-	fwd = tile->stride;
+	tile->w = (tile->w + f-1)>>factor;
+	tile->h = (tile->h + f-1)>>factor;
+	tile->stride = tile->w * (size_t)tile->n;
+	/* Redundant test? We only ever make pixmaps smaller! */
+	if (tile->h > INT_MAX / (tile->w * tile->n))
+		fz_throw(ctx, FZ_ERROR_MEMORY, "pixmap too large");
+	tile->samples = fz_realloc(ctx, tile->samples, (size_t)tile->h * tile->w * tile->n);
+}
+
+void
+fz_subsample_pixblock(unsigned char *s, int w, int h, int n, int factor, ptrdiff_t stride)
+{
+	int fwd, fwd2, fwd3, back, back2, f;
+	unsigned char *d;
+#ifndef ARCH_ARM
+	int x, y, xx, yy, nn;
+#endif
+
+	d = s;
+	f = 1<<factor;
+	fwd = stride;
 	back = f*fwd-n;
 	back2 = f*n-1;
 	fwd2 = (f-1)*n;
-	fwd3 = (f-1)*fwd + tile->stride - w * n;
+	fwd3 = (f-1)*fwd + (int)stride - w * n;
 	factor *= 2;
 #ifdef ARCH_ARM
 	{
@@ -1274,10 +1798,6 @@ fz_subsample_pixmap(fz_context *ctx, fz_pixmap *tile, int factor)
 		}
 	}
 #endif
-	tile->w = dst_w;
-	tile->h = dst_h;
-	tile->stride = dst_w * n;
-	tile->samples = fz_resize_array(ctx, tile->samples, dst_w * n, dst_h);
 }
 
 void
@@ -1287,6 +1807,9 @@ fz_set_pixmap_resolution(fz_context *ctx, fz_pixmap *pix, int xres, int yres)
 	pix->yres = yres;
 }
 
+/*
+	Return the md5 digest for a pixmap
+*/
 void
 fz_md5_pixmap(fz_context *ctx, fz_pixmap *pix, unsigned char digest[16])
 {
@@ -1334,3 +1857,194 @@ int fz_valgrind_pixmap(const fz_pixmap *pix)
 	return total;
 }
 #endif /* HAVE_VALGRIND */
+
+fz_pixmap *
+fz_convert_indexed_pixmap_to_base(fz_context *ctx, const fz_pixmap *src)
+{
+	fz_pixmap *dst;
+	fz_colorspace *base;
+	const unsigned char *s;
+	unsigned char *d;
+	int y, x, k, n, high;
+	unsigned char *lookup;
+	ptrdiff_t s_line_inc, d_line_inc;
+
+	if (src->colorspace->type != FZ_COLORSPACE_INDEXED)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot convert non-indexed pixmap");
+	if (src->n != 1 + src->alpha)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot convert indexed pixmap mis-matching components");
+
+	base = src->colorspace->u.indexed.base;
+	high = src->colorspace->u.indexed.high;
+	lookup = src->colorspace->u.indexed.lookup;
+	n = base->n;
+
+	dst = fz_new_pixmap_with_bbox(ctx, base, fz_pixmap_bbox(ctx, src), src->seps, src->alpha);
+	s = src->samples;
+	d = dst->samples;
+	s_line_inc = src->stride - src->w * (ptrdiff_t)src->n;
+	d_line_inc = dst->stride - dst->w * (ptrdiff_t)dst->n;
+
+	if (src->alpha)
+	{
+		for (y = 0; y < src->h; y++)
+		{
+			for (x = 0; x < src->w; x++)
+			{
+				int v = *s++;
+				int a = *s++;
+				int aa = a + (a>>7);
+				v = fz_mini(v, high);
+				for (k = 0; k < n; k++)
+					*d++ = (aa * lookup[v * n + k] + 128)>>8;
+				*d++ = a;
+			}
+			s += s_line_inc;
+			d += d_line_inc;
+		}
+	}
+	else
+	{
+		for (y = 0; y < src->h; y++)
+		{
+			for (x = 0; x < src->w; x++)
+			{
+				int v = *s++;
+				v = fz_mini(v, high);
+				for (k = 0; k < n; k++)
+					*d++ = lookup[v * n + k];
+			}
+			s += s_line_inc;
+			d += d_line_inc;
+		}
+	}
+
+	if (src->flags & FZ_PIXMAP_FLAG_INTERPOLATE)
+		dst->flags |= FZ_PIXMAP_FLAG_INTERPOLATE;
+	else
+		dst->flags &= ~FZ_PIXMAP_FLAG_INTERPOLATE;
+
+	return dst;
+}
+
+fz_pixmap *
+fz_convert_separation_pixmap_to_base(fz_context *ctx, const fz_pixmap *src)
+{
+	fz_pixmap *dst;
+	fz_colorspace *ss, *base;
+	const unsigned char *s;
+	unsigned char *d;
+	int y, x, k, sn, bn, a;
+	float src_v[FZ_MAX_COLORS];
+	float base_v[FZ_MAX_COLORS];
+	ptrdiff_t s_line_inc, d_line_inc;
+
+	ss = src->colorspace;
+
+	if (ss->type != FZ_COLORSPACE_SEPARATION)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot expand non-separation pixmap");
+	if (src->n != ss->n + src->alpha)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot expand separation pixmap mis-matching alpha channel");
+
+	base = ss->u.separation.base;
+	dst = fz_new_pixmap_with_bbox(ctx, base, fz_pixmap_bbox(ctx, src), src->seps, src->alpha);
+	fz_clear_pixmap(ctx, dst);
+	fz_try(ctx)
+	{
+		s = src->samples;
+		d = dst->samples;
+		s_line_inc = src->stride - src->w * (ptrdiff_t)src->n;
+		d_line_inc = dst->stride - dst->w * (ptrdiff_t)dst->n;
+		sn = ss->n;
+		bn = base->n;
+
+		if (base->type == FZ_COLORSPACE_LAB)
+		{
+			if (src->alpha)
+			{
+				for (y = 0; y < src->h; y++)
+				{
+					for (x = 0; x < src->w; x++)
+					{
+						for (k = 0; k < sn; ++k)
+							src_v[k] = *s++ / 255.0f;
+						a = *s++;
+						ss->u.separation.eval(ctx, ss->u.separation.tint, src_v, sn, base_v, bn);
+						*d++ = (base_v[0] / 100) * 255.0f;
+						*d++ = base_v[1] + 128;
+						*d++ = base_v[2] + 128;
+						*d++ = a;
+					}
+					s += s_line_inc;
+					d += d_line_inc;
+				}
+			}
+			else
+			{
+				for (y = 0; y < src->h; y++)
+				{
+					for (x = 0; x < src->w; x++)
+					{
+						for (k = 0; k < sn; ++k)
+							src_v[k] = *s++ / 255.0f;
+						ss->u.separation.eval(ctx, ss->u.separation.tint, src_v, sn, base_v, bn);
+						*d++ = (base_v[0] / 100) * 255.0f;
+						*d++ = base_v[1] + 128;
+						*d++ = base_v[2] + 128;
+					}
+					s += s_line_inc;
+					d += d_line_inc;
+				}
+			}
+		}
+		else
+		{
+			if (src->alpha)
+			{
+				for (y = 0; y < src->h; y++)
+				{
+					for (x = 0; x < src->w; x++)
+					{
+						for (k = 0; k < sn; ++k)
+							src_v[k] = *s++ / 255.0f;
+						a = *s++;
+						ss->u.separation.eval(ctx, ss->u.separation.tint, src_v, sn, base_v, bn);
+						for (k = 0; k < bn; ++k)
+							*d++ = base_v[k] * 255.0f;
+						*d++ = a;
+					}
+					s += s_line_inc;
+					d += d_line_inc;
+				}
+			}
+			else
+			{
+				for (y = 0; y < src->h; y++)
+				{
+					for (x = 0; x < src->w; x++)
+					{
+						for (k = 0; k < sn; ++k)
+							src_v[k] = *s++ / 255.0f;
+						ss->u.separation.eval(ctx, ss->u.separation.tint, src_v, sn, base_v, bn);
+						for (k = 0; k < bn; ++k)
+							*d++ = base_v[k] * 255.0f;
+					}
+					s += s_line_inc;
+					d += d_line_inc;
+				}
+			}
+		}
+
+		if (src->flags & FZ_PIXMAP_FLAG_INTERPOLATE)
+			dst->flags |= FZ_PIXMAP_FLAG_INTERPOLATE;
+		else
+			dst->flags &= ~FZ_PIXMAP_FLAG_INTERPOLATE;
+	}
+	fz_catch(ctx)
+	{
+		fz_drop_pixmap(ctx, dst);
+		fz_rethrow(ctx);
+	}
+
+	return dst;
+}
