@@ -1,5 +1,29 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
-#include "fitz-imp.h"
+
+#include "color-imp.h"
+#include "pixmap-imp.h"
 
 #include <assert.h>
 #include <string.h>
@@ -9,8 +33,7 @@ enum
 	FZ_SEPARATION_DISABLED_RENDER = 3
 };
 
-
-struct fz_separations_s
+struct fz_separations
 {
 	int refs;
 	int num_separations;
@@ -171,6 +194,37 @@ int fz_count_active_separations(fz_context *ctx, const fz_separations *sep)
 	return c;
 }
 
+int fz_compare_separations(fz_context *ctx, const fz_separations *sep1, const fz_separations *sep2)
+{
+	int i, n1, n2;
+
+	if (sep1 == sep2)
+		return 0; /* Match */
+	if (sep1 == NULL || sep2 == NULL)
+		return 1; /* No match */
+	n1 = sep1->num_separations;
+	n2 = sep2->num_separations;
+	if (n1 != n2)
+		return 1; /* No match */
+	if (sep1->controllable != sep2->controllable)
+		return 1; /* No match */
+	for (i = 0; i < n1; i++)
+	{
+		if (sep_state(sep1, i) != sep_state(sep2, i))
+			return 1; /* No match */
+		if (sep1->name[i] == NULL && sep2->name[i] == NULL)
+		{ /* Two unnamed separations match */ }
+		else if (sep1->name[i] == NULL || sep2->name[i] == NULL || strcmp(sep1->name[i], sep2->name[i]))
+			return 1; /* No match */
+		if (sep1->cs[i] != sep2->cs[i] ||
+			sep1->cs_pos[i] != sep2->cs_pos[i] ||
+			sep1->rgba[i] != sep2->rgba[i] ||
+			sep1->cmyk[i] != sep2->cmyk[i])
+			return 1; /* No match */
+	}
+	return 0;
+}
+
 fz_separations *fz_clone_separations_for_overprint(fz_context *ctx, fz_separations *sep)
 {
 	int i, j, n, c;
@@ -198,11 +252,11 @@ fz_separations *fz_clone_separations_for_overprint(fz_context *ctx, fz_separatio
 	/* We need to clone us a separation structure, with all
 	 * the composite separations marked as enabled. */
 	clone = fz_malloc_struct(ctx, fz_separations);
+	clone->refs = 1;
+	clone->controllable = 0;
 
 	fz_try(ctx)
 	{
-		clone->refs = 1;
-		clone->controllable = 0;
 		for (i = 0; i < n; i++)
 		{
 			fz_separation_behavior beh = sep_state(sep, i);
@@ -227,10 +281,10 @@ fz_separations *fz_clone_separations_for_overprint(fz_context *ctx, fz_separatio
 }
 
 fz_pixmap *
-fz_clone_pixmap_area_with_different_seps(fz_context *ctx, fz_pixmap *src, const fz_irect *bbox, fz_colorspace *dcs, fz_separations *dseps, const fz_color_params *color_params, fz_default_colorspaces *default_cs)
+fz_clone_pixmap_area_with_different_seps(fz_context *ctx, fz_pixmap *src, const fz_irect *bbox, fz_colorspace *dcs, fz_separations *dseps, fz_color_params color_params, fz_default_colorspaces *default_cs)
 {
 	fz_irect local_bbox;
-	fz_pixmap *dst;
+	fz_pixmap *dst, *pix;
 
 	if (bbox == NULL)
 	{
@@ -247,14 +301,19 @@ fz_clone_pixmap_area_with_different_seps(fz_context *ctx, fz_pixmap *src, const
 	else
 		dst->flags &= ~FZ_PIXMAP_FLAG_INTERPOLATE;
 
-	return fz_copy_pixmap_area_converting_seps(ctx, dst, src, color_params, NULL, default_cs);
+	fz_try(ctx)
+		pix = fz_copy_pixmap_area_converting_seps(ctx, src, dst, NULL, color_params, default_cs);
+	fz_catch(ctx)
+	{
+		fz_drop_pixmap(ctx, dst);
+		fz_rethrow(ctx);
+	}
+
+	return pix;
 }
 
-/*
-	We assume that we never map from a DeviceN space to another DeviceN space here.
- */
 fz_pixmap *
-fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *src, const fz_color_params *color_params, fz_colorspace *prf, fz_default_colorspaces *default_cs)
+fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *src, fz_pixmap *dst, fz_colorspace *prf, fz_color_params color_params, fz_default_colorspaces *default_cs)
 {
 	int dw = dst->w;
 	int dh = dst->h;
@@ -274,7 +333,6 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 	int dc = dn - ds - da;
 	const unsigned char *sdata = src->samples + sstride * (dst->y - src->y) + (dst->x - src->x) * sn;
 	unsigned char *ddata = dst->samples;
-	signed char map[FZ_MAX_COLORS];
 	int x, y, i, j, k, n;
 	unsigned char mapped[FZ_MAX_COLORS];
 	int unmapped = sseps_n;
@@ -288,10 +346,14 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 	dstride -= dn * dw;
 	sstride -= sn * dw;
 
+	if (dst->x < src->x || dst->x + dst->w > src->x + src->w ||
+		dst->y < src->y || dst->y + dst->h > src->y + src-> h)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot convert pixmap where dst is not within src!");
+
 	/* Process colorants (and alpha) first */
-	if (dst->colorspace == src->colorspace && proof_cs == NULL)
+	if (dst->colorspace == src->colorspace && proof_cs == NULL && dst->s == 0 && src->s == 0)
 	{
-		/* Simple copy */
+		/* Simple copy - no spots to worry about. */
 		unsigned char *dd = ddata;
 		const unsigned char *sd = sdata;
 		for (y = dh; y > 0; y--)
@@ -369,7 +431,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 		/* Now map the colorants down. */
 		n = fz_colorspace_n(ctx, src->colorspace);
 
-		fz_find_color_converter(ctx, &cc, proof_cs, dst->colorspace, src->colorspace, color_params);
+		fz_find_color_converter(ctx, &cc, src->colorspace, dst->colorspace, proof_cs, color_params);
 
 		fz_try(ctx)
 		{
@@ -382,8 +444,10 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 
 				if (name)
 				{
-					if (!strcmp(name, "None"))
+					if (!strcmp(name, "None")) {
+						mapped[i] = 0;
 						continue;
+					}
 					if (!strcmp(name, "All"))
 					{
 						int n1 = dn - da;
@@ -472,7 +536,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 					/* Src component i is not mapped. We need to convert that down. */
 					memset(colors, 0, sizeof(float) * n);
 					colors[i] = 1;
-					cc.convert(ctx, &cc, convert, colors);
+					cc.convert(ctx, &cc, colors, convert);
 
 					if (fz_colorspace_is_subtractive(ctx, dst->colorspace))
 					{
@@ -579,7 +643,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 
 						memset(colors, 0, sizeof(float) * n);
 						colors[i] = 1;
-						cc.convert(ctx, &cc, convert, colors);
+						cc.convert(ctx, &cc, colors, convert);
 
 						if (sa)
 						{
@@ -632,7 +696,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 							{
 								for (j = 0; j < n; j++)
 									colors[j] = mapped[j] ? 0 : sd[j] / 255.0f;
-								cc.convert(ctx, &cc, convert, colors);
+								cc.convert(ctx, &cc, colors, convert);
 
 								for (j = 0; j < dc; j++)
 									dd[j] = fz_clampi(255 * convert[j], 0, 255);
@@ -650,13 +714,18 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 							for (x = dw; x > 0; x--)
 							{
 								unsigned char a = sd[sc];
-								float inva = 1.0f/a;
-								for (j = 0; j < n; j++)
-									colors[j] = mapped[j] ? 0 : sd[j] * inva;
-								cc.convert(ctx, &cc, convert, colors);
+								if (a == 0)
+									memset(dd, 0, dc);
+								else
+								{
+									float inva = 1.0f/a;
+									for (j = 0; j < n; j++)
+										colors[j] = mapped[j] ? 0 : sd[j] * inva;
+									cc.convert(ctx, &cc, colors, convert);
 
-								for (j = 0; j < dc; j++)
-									dd[j] = fz_clampi(a * convert[j], 0, a);
+									for (j = 0; j < dc; j++)
+										dd[j] = fz_clampi(a * convert[j], 0, a);
+								}
 								dd += dn;
 								sd += sn;
 							}
@@ -675,29 +744,65 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 	}
 	else
 	{
+		signed char map[FZ_MAX_COLORS];
+
+		/* We have a special case here. Converting from CMYK + Spots
+		 * to RGB with less spots, involves folding (at least some of)
+		 * the spots down via their equivalent colors. Merging a spot's
+		 * equivalent colour (generally expressed in CMYK) with an RGB
+		 * one works badly, (presumably because RGB colors have
+		 * different linearity to CMYK ones). For best results we want
+		 * to merge the spots into the CMYK color, and then convert
+		 * that into RGB.  We handle that case here. */
+		if (fz_colorspace_is_subtractive(ctx, src->colorspace) &&
+			!fz_colorspace_is_subtractive(ctx, dst->colorspace) &&
+			src->seps > 0 &&
+			fz_compare_separations(ctx, dst->seps, src->seps))
+		{
+			/* Converting from CMYK + Spots -> RGB with a change in spots. */
+			fz_pixmap *temp = fz_new_pixmap(ctx, src->colorspace, src->w, src->h, dst->seps, dst->alpha);
+
+			/* Match the regions exactly (this matters in particular when we are
+			 * using rotation, and the src region is not origined at 0,0 - see bug
+			 * 704726. */
+			temp->x = src->x;
+			temp->y = src->y;
+
+			fz_try(ctx)
+			{
+				temp = fz_copy_pixmap_area_converting_seps(ctx, src, temp, prf, color_params, default_cs);
+				dst =  fz_copy_pixmap_area_converting_seps(ctx, temp, dst, NULL, color_params, default_cs);
+			}
+			fz_always(ctx)
+				fz_drop_pixmap(ctx, temp);
+			fz_catch(ctx)
+				fz_rethrow(ctx);
+
+			return dst;
+		}
+
 		/* Use a standard pixmap converter to convert the process + alpha. */
-		fz_pixmap_converter *pc = fz_lookup_pixmap_converter(ctx, dst->colorspace, src->colorspace);
-		pc(ctx, dst, src, proof_cs, default_cs, NULL, 0);
+		fz_convert_pixmap_samples(ctx, src, dst, proof_cs, default_cs, fz_default_color_params, 0);
 
 		/* And handle the spots ourselves. First make a map of what spots go where. */
 		/* We want to set it up so that:
 		 *    For each source spot, i, mapped[i] != 0 implies that it maps directly to a dest spot.
-		 *    For each dest spot, j, mapped[j] = the source spot that goes there (or -1 if none).
+		 *    For each dest spot, j, map[j] = the source spot that goes there (or -1 if none).
 		 */
 		for (i = 0; i < sseps_n; i++)
 			mapped[i] = 0;
 
-		for (i = 0, k = 0; i < dseps_n; i++)
+		for (i = 0; i < dseps_n; i++)
 		{
 			const char *name;
 			int state = sep_state(dseps, i);
 
+			map[i] = -1;
 			if (state != FZ_SEPARATION_SPOT)
 				continue;
 			name = dseps->name[i];
 			if (name == NULL)
 				continue;
-			map[k] = -1;
 			for (j = 0; j < sseps_n; j++)
 			{
 				const char *sname;
@@ -708,16 +813,16 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 				sname = sseps->name[j];
 				if (sname && !strcmp(name, sname))
 				{
-					map[k] = j;
+					map[i] = j;
 					unmapped--;
 					mapped[j] = 1;
 					break;
 				}
 			}
-			k++;
 		}
 		if (sa)
-			map[k] = sseps_n;
+			map[i] = sseps_n;
+		/* map[i] is now defined for all 0 <= i < dseps_n+sa */
 
 		/* Now we need to make d[i] = map[i] < 0 : 0 ? s[map[i]] */
 		if (ds)
@@ -755,7 +860,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 					continue;
 				i++;
 				/* Src spot m (the i'th one) is not mapped. We need to convert that down. */
-				fz_separation_equivalent(ctx, sseps, m, color_params, dst->colorspace, proof_cs, convert);
+				fz_separation_equivalent(ctx, sseps, m, dst->colorspace, convert, proof_cs, color_params);
 
 				if (fz_colorspace_is_subtractive(ctx, dst->colorspace))
 				{
@@ -763,6 +868,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 					{
 						unsigned char *dd = ddata;
 						const unsigned char *sd = sdata + sc;
+
 						if (sa)
 						{
 							for (y = dh; y > 0; y--)
@@ -772,7 +878,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 									unsigned char v = sd[i];
 									if (v != 0)
 									{
-										unsigned char a = sd[sc];
+										unsigned char a = sd[ss];
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] + v * convert[k], 0, a);
 									}
@@ -785,6 +891,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 						}
 						else
 						{
+							/* This case is exercised by: -o out%d.pgm -r72 -D -F pgm -stm ../perf-testing-gpdl/pdf/Ad_InDesign.pdf */
 							for (y = dh; y > 0; y--)
 							{
 								for (x = dw; x > 0; x--)
@@ -805,16 +912,17 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 					{
 						unsigned char *dd = ddata;
 						const unsigned char *sd = sdata + sc;
+
 						if (sa)
 						{
 							for (y = dh; y > 0; y--)
 							{
 								for (x = dw; x > 0; x--)
 								{
-									unsigned char v = 0xff - sd[i];
+									unsigned char v = sd[i];
 									if (v != 0)
 									{
-										unsigned char a = sd[sc];
+										unsigned char a = sd[ss];
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] + v * convert[k], 0, a);
 									}
@@ -827,11 +935,12 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 						}
 						else
 						{
+							/* This case is exercised by: -o out.pkm -r72 -D ../MyTests/Bug704778.pdf 1 */
 							for (y = dh; y > 0; y--)
 							{
 								for (x = dw; x > 0; x--)
 								{
-									unsigned char v = 0xff - sd[i];
+									unsigned char v = sd[i];
 									if (v != 0)
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] + v * convert[k], 0, 255);
@@ -852,6 +961,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 					{
 						unsigned char *dd = ddata;
 						const unsigned char *sd = sdata + sc;
+
 						if (sa)
 						{
 							for (y = dh; y > 0; y--)
@@ -861,7 +971,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 									unsigned char v = sd[i];
 									if (v != 0)
 									{
-										unsigned char a = sd[sc];
+										unsigned char a = sd[ss];
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] - v * convert[k], 0, a);
 									}
@@ -874,6 +984,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 						}
 						else
 						{
+							/* Nothing in the cluster tests this case. */
 							for (y = dh; y > 0; y--)
 							{
 								for (x = dw; x > 0; x--)
@@ -894,16 +1005,17 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 					{
 						unsigned char *dd = ddata;
 						const unsigned char *sd = sdata + sc;
+
 						if (sa)
 						{
 							for (y = dh; y > 0; y--)
 							{
 								for (x = dw; x > 0; x--)
 								{
-									unsigned char v = 0xff - sd[i];
+									unsigned char v = sd[i];
 									if (v != 0)
 									{
-										unsigned char a = sd[sc];
+										unsigned char a = sd[ss];
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] - v * convert[k], 0, a);
 									}
@@ -916,11 +1028,12 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 						}
 						else
 						{
+							/* This case is exercised by: -o out.png -r72 -D ../MyTests/Bug704778.pdf 1 */
 							for (y = dh; y > 0; y--)
 							{
 								for (x = dw; x > 0; x--)
 								{
-									unsigned char v = 0xff - sd[i];
+									unsigned char v = sd[i];
 									if (v != 0)
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] - v * convert[k], 0, 255);
@@ -940,7 +1053,11 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 	return dst;
 }
 
-void fz_convert_separation_colors(fz_context *ctx, const fz_color_params *color_params, const fz_colorspace *dst_cs, const fz_separations *dst_seps, float *dst_color, const fz_colorspace *src_cs, const float *src_color)
+void
+fz_convert_separation_colors(fz_context *ctx,
+	fz_colorspace *src_cs, const float *src_color,
+	fz_separations *dst_seps, fz_colorspace *dst_cs, float *dst_color,
+	fz_color_params color_params)
 {
 	int i, j, n, dc, ds, dn, pred;
 	float remainders[FZ_MAX_COLORS];
@@ -1025,14 +1142,19 @@ found_process:
 	{
 		/* There were some spots that didn't copy over */
 		float converted[FZ_MAX_COLORS];
-		fz_convert_color(ctx, color_params, NULL, dst_cs, converted, src_cs, remainders);
-
+		fz_convert_color(ctx, src_cs, remainders, dst_cs, converted, NULL, color_params);
 		for (i = 0; i < dc; i++)
 			dst_color[i] += converted[i];
 	}
 }
 
-void fz_separation_equivalent(fz_context *ctx, const fz_separations *seps, int i, const fz_color_params *color_params, const fz_colorspace *dst_cs, const fz_colorspace *prf, float *convert)
+void
+fz_separation_equivalent(fz_context *ctx,
+	const fz_separations *seps,
+	int i,
+	fz_colorspace *dst_cs, float *convert,
+	fz_colorspace *prf,
+	fz_color_params color_params)
 {
 	float colors[FZ_MAX_COLORS];
 
@@ -1059,5 +1181,5 @@ void fz_separation_equivalent(fz_context *ctx, const fz_separations *seps, int i
 
 	memset(colors, 0, sizeof(float) * fz_colorspace_n(ctx, seps->cs[i]));
 	colors[seps->cs_pos[i]] = 1;
-	fz_convert_color(ctx, color_params, prf, dst_cs, convert, seps->cs[i], colors);
+	fz_convert_color(ctx, seps->cs[i], colors, dst_cs, convert, prf, color_params);
 }
