@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
@@ -6,6 +28,8 @@
 #include <float.h>
 #include <limits.h>
 
+static pdf_function *pdf_load_function_imp(fz_context *ctx, pdf_obj *dict, int in, int out, pdf_cycle_list *cycle);
+
 #define DIV_BY_ZERO(a, b, min, max) (((a) < 0) ^ ((b) < 0) ? (min) : (max))
 
 enum
@@ -14,8 +38,6 @@ enum
 	MAX_M = FZ_MAX_COLORS
 };
 
-typedef struct psobj_s psobj;
-
 enum
 {
 	SAMPLE = 0,
@@ -24,7 +46,20 @@ enum
 	POSTSCRIPT = 4
 };
 
-struct pdf_function_s
+typedef struct
+{
+	int type;
+	union
+	{
+		int b;				/* boolean (stack only) */
+		int i;				/* integer (stack and code) */
+		float f;			/* real (stack and code) */
+		int op;				/* operator (code only) */
+		int block;			/* if/ifelse block pointer (code only) */
+	} u;
+} psobj;
+
+struct pdf_function
 {
 	fz_storable storable;
 	size_t size;
@@ -120,57 +155,11 @@ static char *ps_op_names[] =
 	"roll", "round", "sin", "sqrt", "sub", "true", "truncate", "xor"
 };
 
-struct psobj_s
-{
-	int type;
-	union
-	{
-		int b;				/* boolean (stack only) */
-		int i;				/* integer (stack and code) */
-		float f;			/* real (stack and code) */
-		int op;				/* operator (code only) */
-		int block;			/* if/ifelse block pointer (code only) */
-	} u;
-};
-
-typedef struct ps_stack_s ps_stack;
-
-struct ps_stack_s
+typedef struct
 {
 	psobj stack[100];
 	int sp;
-};
-
-void
-pdf_print_ps_stack(fz_context *ctx, fz_output *out, ps_stack *st)
-{
-	int i;
-
-	fz_write_printf(ctx, out, "stack:");
-
-	for (i = 0; i < st->sp; i++)
-	{
-		switch (st->stack[i].type)
-		{
-		case PS_BOOL:
-			if (st->stack[i].u.b)
-				fz_write_printf(ctx, out, " true");
-			else
-				fz_write_printf(ctx, out, " false");
-			break;
-
-		case PS_INT:
-			fz_write_printf(ctx, out, " %d", st->stack[i].u.i);
-			break;
-
-		case PS_REAL:
-			fz_write_printf(ctx, out, " %g", st->stack[i].u.f);
-			break;
-		}
-	}
-
-	fz_write_printf(ctx, out, "\n");
-}
+} ps_stack;
 
 static void
 ps_init_stack(ps_stack *st)
@@ -181,7 +170,7 @@ ps_init_stack(ps_stack *st)
 
 static inline int ps_overflow(ps_stack *st, int n)
 {
-	return n < 0 || st->sp + n >= nelem(st->stack);
+	return n < 0 || st->sp + n >= (int)nelem(st->stack);
 }
 
 static inline int ps_underflow(ps_stack *st, int n)
@@ -390,7 +379,7 @@ ps_run(fz_context *ctx, psobj *code, ps_stack *st, int pc)
 			case PS_OP_BITSHIFT:
 				i2 = ps_pop_int(st);
 				i1 = ps_pop_int(st);
-				if (i2 > 0 && i2 < 8 * sizeof (i2))
+				if (i2 > 0 && i2 < 8 * (int)sizeof (i2))
 					ps_push_int(st, i1 << i2);
 				else if (i2 < 0 && i2 > -8 * (int)sizeof (i2))
 					ps_push_int(st, (int)((unsigned int)i1 >> -i2));
@@ -717,18 +706,21 @@ resize_code(fz_context *ctx, pdf_function *func, int newsize)
 	if (newsize >= func->u.p.cap)
 	{
 		int new_cap = func->u.p.cap + 64;
-		func->u.p.code = fz_resize_array(ctx, func->u.p.code, new_cap, sizeof(psobj));
+		func->u.p.code = fz_realloc_array(ctx, func->u.p.code, new_cap, psobj);
 		func->u.p.cap = new_cap;
 	}
 }
 
 static void
-parse_code(fz_context *ctx, pdf_function *func, fz_stream *stream, int *codeptr, pdf_lexbuf *buf)
+parse_code(fz_context *ctx, pdf_function *func, fz_stream *stream, int *codeptr, pdf_lexbuf *buf, int depth)
 {
 	pdf_token tok;
 	int opptr, elseptr, ifptr;
 	int a, b, mid, cmp;
 
+	if (depth > 100)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "too much recursion in calculator function");
+
 	while (1)
 	{
 		tok = pdf_lex(ctx, stream, buf);
@@ -773,14 +765,14 @@ parse_code(fz_context *ctx, pdf_function *func, fz_stream *stream, int *codeptr,
 			resize_code(ctx, func, *codeptr);
 
 			ifptr = *codeptr;
-			parse_code(ctx, func, stream, codeptr, buf);
+			parse_code(ctx, func, stream, codeptr, buf, depth + 1);
 
 			tok = pdf_lex(ctx, stream, buf);
 
 			if (tok == PDF_TOK_OPEN_BRACE)
 			{
 				elseptr = *codeptr;
-				parse_code(ctx, func, stream, codeptr, buf);
+				parse_code(ctx, func, stream, codeptr, buf, depth + 1);
 
 				tok = pdf_lex(ctx, stream, buf);
 			}
@@ -870,12 +862,10 @@ load_postscript_func(fz_context *ctx, pdf_function *func, pdf_obj *dict)
 	int codeptr;
 	pdf_lexbuf buf;
 	pdf_token tok;
-	int locked = 0;
 
 	pdf_lexbuf_init(ctx, &buf, PDF_LEXBUF_SMALL);
 
 	fz_var(stream);
-	fz_var(locked);
 
 	fz_try(ctx)
 	{
@@ -891,7 +881,7 @@ load_postscript_func(fz_context *ctx, pdf_function *func, pdf_obj *dict)
 		func->u.p.cap = 0;
 
 		codeptr = 0;
-		parse_code(ctx, func, stream, &codeptr, &buf);
+		parse_code(ctx, func, stream, &codeptr, &buf, 0);
 	}
 	fz_always(ctx)
 	{
@@ -945,8 +935,6 @@ load_sample_func(fz_context *ctx, pdf_function *func, pdf_obj *dict)
 	int bps;
 	int i;
 
-	fz_var(stream);
-
 	func->u.sa.samples = NULL;
 
 	obj = pdf_dict_get(ctx, dict, PDF_NAME(Size));
@@ -1012,7 +1000,7 @@ load_sample_func(fz_context *ctx, pdf_function *func, pdf_obj *dict)
 	if (samplecount > MAX_SAMPLE_FUNCTION_SIZE)
 		fz_throw(ctx, FZ_ERROR_SYNTAX, "sample function too large");
 
-	func->u.sa.samples = fz_malloc_array(ctx, samplecount, sizeof(float));
+	func->u.sa.samples = Memento_label(fz_malloc_array(ctx, samplecount, float), "function_samples");
 	func->size += samplecount * sizeof(float);
 
 	stream = pdf_open_stream(ctx, dict);
@@ -1093,7 +1081,7 @@ eval_sample_func(fz_context *ctx, pdf_function *func, const float *in, float *ou
 		x = fz_clamp(x, 0, func->u.sa.size[i] - 1);
 		e0[i] = floorf(x);
 		e1[i] = ceilf(x);
-		efrac[i] = x - floorf(x);
+		efrac[i] = x - e0[i];
 	}
 
 	scale[0] = func->n;
@@ -1214,7 +1202,11 @@ eval_exponential_func(fz_context *ctx, pdf_function *func, float in, float *out)
 
 	/* Default output is zero, which is suitable for violated constraints */
 	if ((func->u.e.n != (int)func->u.e.n && x < 0) || (func->u.e.n < 0 && x == 0))
+	{
+		for (i = 0; i < func->n; i++)
+			out[i] = 0;
 		return;
+	}
 
 	tmp = powf(x, func->u.e.n);
 	for (i = 0; i < func->n; i++)
@@ -1230,7 +1222,7 @@ eval_exponential_func(fz_context *ctx, pdf_function *func, float in, float *out)
  */
 
 static void
-load_stitching_func(fz_context *ctx, pdf_function *func, pdf_obj *dict)
+load_stitching_func(fz_context *ctx, pdf_function *func, pdf_obj *dict, pdf_cycle_list *cycle_up)
 {
 	pdf_function **funcs;
 	pdf_obj *obj;
@@ -1249,38 +1241,25 @@ load_stitching_func(fz_context *ctx, pdf_function *func, pdf_obj *dict)
 	if (!pdf_is_array(ctx, obj))
 		fz_throw(ctx, FZ_ERROR_SYNTAX, "stitching function has no input functions");
 
-	fz_try(ctx)
+	k = pdf_array_len(ctx, obj);
+
+	func->u.st.funcs = Memento_label(fz_malloc_array(ctx, k, pdf_function*), "stitch_fns");
+	func->u.st.bounds = Memento_label(fz_malloc_array(ctx, k - 1, float), "stitch_bounds");
+	func->u.st.encode = Memento_label(fz_malloc_array(ctx, k * 2, float), "stitch_encode");
+	funcs = func->u.st.funcs;
+
+	for (i = 0; i < k; i++)
 	{
-		if (pdf_mark_obj(ctx, obj))
-			fz_throw(ctx, FZ_ERROR_SYNTAX, "recursive function");
-		k = pdf_array_len(ctx, obj);
+		sub = pdf_array_get(ctx, obj, i);
+		funcs[i] = pdf_load_function_imp(ctx, sub, 1, func->n, cycle_up);
 
-		func->u.st.funcs = fz_malloc_array(ctx, k, sizeof(pdf_function*));
-		func->u.st.bounds = fz_malloc_array(ctx, k - 1, sizeof(float));
-		func->u.st.encode = fz_malloc_array(ctx, k * 2, sizeof(float));
-		funcs = func->u.st.funcs;
+		func->size += pdf_function_size(ctx, funcs[i]);
+		func->u.st.k ++;
 
-		for (i = 0; i < k; i++)
-		{
-			sub = pdf_array_get(ctx, obj, i);
-			funcs[i] = pdf_load_function(ctx, sub, 1, func->n);
-
-			func->size += pdf_function_size(ctx, funcs[i]);
-			func->u.st.k ++;
-
-			if (funcs[i]->m != func->m)
-				fz_warn(ctx, "wrong number of inputs for sub function %d", i);
-			if (funcs[i]->n != func->n)
-				fz_warn(ctx, "wrong number of outputs for sub function %d", i);
-		}
-	}
-	fz_always(ctx)
-	{
-		pdf_unmark_obj(ctx, obj);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
+		if (funcs[i]->m != func->m)
+			fz_warn(ctx, "wrong number of inputs for sub function %d", i);
+		if (funcs[i]->n != func->n)
+			fz_warn(ctx, "wrong number of outputs for sub function %d", i);
 	}
 
 	obj = pdf_dict_get(ctx, dict, PDF_NAME(Bounds));
@@ -1365,7 +1344,7 @@ eval_stitching_func(fz_context *ctx, pdf_function *func, float in, float *out)
 
 	in = lerp(in, low, high, func->u.st.encode[i * 2 + 0], func->u.st.encode[i * 2 + 1]);
 
-	pdf_eval_function(ctx, func->u.st.funcs[i], &in, 1, out, func->u.st.funcs[i]->n);
+	pdf_eval_function(ctx, func->u.st.funcs[i], &in, 1, out, func->n);
 }
 
 /*
@@ -1420,9 +1399,9 @@ pdf_eval_function(fz_context *ctx, pdf_function *func, const float *in, int inle
 
 	if (inlen < func->m)
 	{
-		for (i = 0; i < func->m; ++i)
+		for (i = 0; i < inlen; ++i)
 			fakein[i] = in[i];
-		for (; i < inlen; ++i)
+		for (; i < func->m; ++i)
 			fakein[i] = 0;
 		in = fakein;
 	}
@@ -1441,15 +1420,16 @@ pdf_eval_function(fz_context *ctx, pdf_function *func, const float *in, int inle
 	}
 }
 
-pdf_function *
-pdf_load_function(fz_context *ctx, pdf_obj *dict, int in, int out)
+static pdf_function *
+pdf_load_function_imp(fz_context *ctx, pdf_obj *dict, int in, int out, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	pdf_function *func;
 	pdf_obj *obj;
 	int i;
 
-	if (pdf_obj_marked(ctx, dict))
-		fz_throw(ctx, FZ_ERROR_SYNTAX, "Recursion in function definition");
+	if (pdf_cycle(ctx, &cycle, cycle_up, dict))
+		fz_throw(ctx, FZ_ERROR_SYNTAX, "recursive function");
 
 	if ((func = pdf_find_item(ctx, pdf_drop_function_imp, dict)) != NULL)
 		return func;
@@ -1506,7 +1486,7 @@ pdf_load_function(fz_context *ctx, pdf_obj *dict, int in, int out)
 			break;
 
 		case STITCHING:
-			load_stitching_func(ctx, func, dict);
+			load_stitching_func(ctx, func, dict, &cycle);
 			break;
 
 		case POSTSCRIPT:
@@ -1527,3 +1507,9 @@ pdf_load_function(fz_context *ctx, pdf_obj *dict, int in, int out)
 
 	return func;
 }
+
+pdf_function *
+pdf_load_function(fz_context *ctx, pdf_obj *dict, int in, int out)
+{
+	return pdf_load_function_imp(ctx, dict, in, out, NULL);
+}
