@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 /*
  * Information tool.
  * Print information about the input pdf.
@@ -70,7 +92,7 @@ struct info
 	} u;
 };
 
-typedef struct globals_s
+typedef struct
 {
 	pdf_document *doc;
 	fz_context *ctx;
@@ -173,7 +195,6 @@ infousage(void)
 		"\t-X\tlist form and postscript xobjects\n"
 		"\tpages\tcomma separated list of page numbers and ranges\n"
 		);
-	exit(1);
 }
 
 static void
@@ -182,21 +203,22 @@ showglobalinfo(fz_context *ctx, globals *glo)
 	pdf_obj *obj;
 	fz_output *out = glo->out;
 	pdf_document *doc = glo->doc;
+	int version = pdf_version(ctx, doc);
 
-	fz_write_printf(ctx, out, "\nPDF-%d.%d\n", doc->version / 10, doc->version % 10);
+	fz_write_printf(ctx, out, "\nPDF-%d.%d\n", version / 10, version % 10);
 
 	obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Info));
 	if (obj)
 	{
 		fz_write_printf(ctx, out, "Info object (%d 0 R):\n", pdf_to_num(ctx, obj));
-		pdf_print_obj(ctx, out, pdf_resolve_indirect(ctx, obj), 1);
+		pdf_print_obj(ctx, out, pdf_resolve_indirect(ctx, obj), 1, 1);
 	}
 
 	obj = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Encrypt));
 	if (obj)
 	{
 		fz_write_printf(ctx, out, "\nEncryption object (%d 0 R):\n", pdf_to_num(ctx, obj));
-		pdf_print_obj(ctx, out, pdf_resolve_indirect(ctx, obj), 1);
+		pdf_print_obj(ctx, out, pdf_resolve_indirect(ctx, obj), 1, 1);
 	}
 
 	fz_write_printf(ctx, out, "\nPages: %d\n\n", glo->pagecount);
@@ -216,7 +238,7 @@ gatherdimensions(fz_context *ctx, globals *glo, int page, pdf_obj *pageref)
 	bbox = pdf_to_rect(ctx, obj);
 
 	obj = pdf_dict_get(ctx, pageref, PDF_NAME(UserUnit));
-	if (pdf_is_real(ctx, obj))
+	if (pdf_is_number(ctx, obj))
 	{
 		float unit = pdf_to_real(ctx, obj);
 		bbox.x0 *= unit;
@@ -232,11 +254,12 @@ gatherdimensions(fz_context *ctx, globals *glo, int page, pdf_obj *pageref)
 	if (j < glo->dims)
 		return;
 
-	glo->dim = fz_resize_array(ctx, glo->dim, glo->dims+1, sizeof(struct info));
+	glo->dim = fz_realloc_array(ctx, glo->dim, glo->dims+1, struct info);
 	glo->dims++;
 
 	glo->dim[glo->dims - 1].page = page;
 	glo->dim[glo->dims - 1].pageref = pageref;
+	glo->dim[glo->dims - 1].u.dim.bbox = NULL;
 	glo->dim[glo->dims - 1].u.dim.bbox = fz_malloc(ctx, sizeof(fz_rect));
 	memcpy(glo->dim[glo->dims - 1].u.dim.bbox, &bbox, sizeof (fz_rect));
 
@@ -280,7 +303,7 @@ gatherfonts(fz_context *ctx, globals *glo, int page, pdf_obj *pageref, pdf_obj *
 		if (k < glo->fonts)
 			continue;
 
-		glo->font = fz_resize_array(ctx, glo->font, glo->fonts+1, sizeof(struct info));
+		glo->font = fz_realloc_array(ctx, glo->font, glo->fonts+1, struct info);
 		glo->fonts++;
 
 		glo->font[glo->fonts - 1].page = page;
@@ -349,7 +372,7 @@ gatherimages(fz_context *ctx, globals *glo, int page, pdf_obj *pageref, pdf_obj
 		if (k < glo->images)
 			continue;
 
-		glo->image = fz_resize_array(ctx, glo->image, glo->images+1, sizeof(struct info));
+		glo->image = fz_realloc_array(ctx, glo->image, glo->images+1, struct info);
 		glo->images++;
 
 		glo->image[glo->images - 1].page = page;
@@ -406,7 +429,7 @@ gatherforms(fz_context *ctx, globals *glo, int page, pdf_obj *pageref, pdf_obj *
 		if (k < glo->forms)
 			continue;
 
-		glo->form = fz_resize_array(ctx, glo->form, glo->forms+1, sizeof(struct info));
+		glo->form = fz_realloc_array(ctx, glo->form, glo->forms+1, struct info);
 		glo->forms++;
 
 		glo->form[glo->forms - 1].page = page;
@@ -450,7 +473,7 @@ gatherpsobjs(fz_context *ctx, globals *glo, int page, pdf_obj *pageref, pdf_obj
 		if (k < glo->psobjs)
 			continue;
 
-		glo->psobj = fz_resize_array(ctx, glo->psobj, glo->psobjs+1, sizeof(struct info));
+		glo->psobj = fz_realloc_array(ctx, glo->psobj, glo->psobjs+1, struct info);
 		glo->psobjs++;
 
 		glo->psobj[glo->psobjs - 1].page = page;
@@ -492,7 +515,7 @@ gathershadings(fz_context *ctx, globals *glo, int page, pdf_obj *pageref, pdf_ob
 		if (k < glo->shadings)
 			continue;
 
-		glo->shading = fz_resize_array(ctx, glo->shading, glo->shadings+1, sizeof(struct info));
+		glo->shading = fz_realloc_array(ctx, glo->shading, glo->shadings+1, struct info);
 		glo->shadings++;
 
 		glo->shading[glo->shadings - 1].page = page;
@@ -559,7 +582,7 @@ gatherpatterns(fz_context *ctx, globals *glo, int page, pdf_obj *pageref, pdf_ob
 		if (k < glo->patterns)
 			continue;
 
-		glo->pattern = fz_resize_array(ctx, glo->pattern, glo->patterns+1, sizeof(struct info));
+		glo->pattern = fz_realloc_array(ctx, glo->pattern, glo->patterns+1, struct info);
 		glo->patterns++;
 
 		glo->pattern[glo->patterns - 1].page = page;
@@ -573,95 +596,80 @@ gatherpatterns(fz_context *ctx, globals *glo, int page, pdf_obj *pageref, pdf_ob
 }
 
 static void
-gatherresourceinfo(fz_context *ctx, globals *glo, int page, pdf_obj *rsrc, int show)
+gatherresourceinfo(fz_context *ctx, pdf_mark_list *mark_list, globals *glo, int page, pdf_obj *obj, int show)
 {
+	pdf_obj *rsrc;
 	pdf_obj *pageref;
 	pdf_obj *font;
 	pdf_obj *xobj;
 	pdf_obj *shade;
 	pdf_obj *pattern;
-	pdf_obj *subrsrc;
 	int i;
 
+	/* stop on cyclic resource dependencies */
+	if (pdf_mark_list_push(ctx, mark_list, obj))
+		return;
+
+	rsrc = pdf_dict_get(ctx, obj, PDF_NAME(Resources));
+
 	pageref = pdf_lookup_page_obj(ctx, glo->doc, page-1);
 	if (!pageref)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot retrieve info from page %d", page);
 
-	/* stop on cyclic resource dependencies */
-	if (pdf_mark_obj(ctx, rsrc))
-		return;
-
-	fz_try(ctx)
+	font = pdf_dict_get(ctx, rsrc, PDF_NAME(Font));
+	if (show & FONTS && font && !pdf_mark_list_push(ctx, mark_list, font))
 	{
-		font = pdf_dict_get(ctx, rsrc, PDF_NAME(Font));
-		if (show & FONTS && font)
+		int n;
+
+		gatherfonts(ctx, glo, page, pageref, font);
+		n = pdf_dict_len(ctx, font);
+		for (i = 0; i < n; i++)
 		{
-			int n;
+			gatherresourceinfo(ctx, mark_list, glo, page, pdf_dict_get_val(ctx, font, i), show);
+		}
+	}
 
-			gatherfonts(ctx, glo, page, pageref, font);
-			n = pdf_dict_len(ctx, font);
-			for (i = 0; i < n; i++)
-			{
-				pdf_obj *obj = pdf_dict_get_val(ctx, font, i);
+	xobj = pdf_dict_get(ctx, rsrc, PDF_NAME(XObject));
+	if (show & (IMAGES|XOBJS) && xobj && !pdf_mark_list_push(ctx, mark_list, xobj))
+	{
+		int n;
 
-				subrsrc = pdf_dict_get(ctx, obj, PDF_NAME(Resources));
-				if (subrsrc && pdf_objcmp(ctx, rsrc, subrsrc))
-					gatherresourceinfo(ctx, glo, page, subrsrc, show);
-			}
+		if (show & IMAGES)
+			gatherimages(ctx, glo, page, pageref, xobj);
+		if (show & XOBJS)
+		{
+			gatherforms(ctx, glo, page, pageref, xobj);
+			gatherpsobjs(ctx, glo, page, pageref, xobj);
 		}
-
-		xobj = pdf_dict_get(ctx, rsrc, PDF_NAME(XObject));
-		if (show & (IMAGES|XOBJS) && xobj)
+		n = pdf_dict_len(ctx, xobj);
+		for (i = 0; i < n; i++)
 		{
-			int n;
-
-			if (show & IMAGES)
-				gatherimages(ctx, glo, page, pageref, xobj);
-			if (show & XOBJS)
-			{
-				gatherforms(ctx, glo, page, pageref, xobj);
-				gatherpsobjs(ctx, glo, page, pageref, xobj);
-			}
-			n = pdf_dict_len(ctx, xobj);
-			for (i = 0; i < n; i++)
-			{
-				pdf_obj *obj = pdf_dict_get_val(ctx, xobj, i);
-				subrsrc = pdf_dict_get(ctx, obj, PDF_NAME(Resources));
-				if (subrsrc && pdf_objcmp(ctx, rsrc, subrsrc))
-					gatherresourceinfo(ctx, glo, page, subrsrc, show);
-			}
+			gatherresourceinfo(ctx, mark_list, glo, page, pdf_dict_get_val(ctx, xobj, i), show);
 		}
+	}
 
-		shade = pdf_dict_get(ctx, rsrc, PDF_NAME(Shading));
-		if (show & SHADINGS && shade)
-			gathershadings(ctx, glo, page, pageref, shade);
+	shade = pdf_dict_get(ctx, rsrc, PDF_NAME(Shading));
+	if (show & SHADINGS && shade && !pdf_mark_list_push(ctx, mark_list, shade))
+		gathershadings(ctx, glo, page, pageref, shade);
 
-		pattern = pdf_dict_get(ctx, rsrc, PDF_NAME(Pattern));
-		if (show & PATTERNS && pattern)
+	pattern = pdf_dict_get(ctx, rsrc, PDF_NAME(Pattern));
+	if (show & PATTERNS && pattern && !pdf_mark_list_push(ctx, mark_list, pattern))
+	{
+		int n;
+		gatherpatterns(ctx, glo, page, pageref, pattern);
+		n = pdf_dict_len(ctx, pattern);
+		for (i = 0; i < n; i++)
 		{
-			int n;
-			gatherpatterns(ctx, glo, page, pageref, pattern);
-			n = pdf_dict_len(ctx, pattern);
-			for (i = 0; i < n; i++)
-			{
-				pdf_obj *obj = pdf_dict_get_val(ctx, pattern, i);
-				subrsrc = pdf_dict_get(ctx, obj, PDF_NAME(Resources));
-				if (subrsrc && pdf_objcmp(ctx, rsrc, subrsrc))
-					gatherresourceinfo(ctx, glo, page, subrsrc, show);
-			}
+			gatherresourceinfo(ctx, mark_list, glo, page, pdf_dict_get_val(ctx, pattern, i), show);
 		}
 	}
-	fz_always(ctx)
-		pdf_unmark_obj(ctx, rsrc);
-	fz_catch(ctx)
-		fz_rethrow(ctx);
 }
 
 static void
 gatherpageinfo(fz_context *ctx, globals *glo, int page, int show)
 {
+	pdf_mark_list mark_list;
 	pdf_obj *pageref;
-	pdf_obj *rsrc;
 
 	pageref = pdf_lookup_page_obj(ctx, glo->doc, page-1);
 
@@ -670,8 +678,13 @@ gatherpageinfo(fz_context *ctx, globals *glo, int page, int show)
 
 	gatherdimensions(ctx, glo, page, pageref);
 
-	rsrc = pdf_dict_get(ctx, pageref, PDF_NAME(Resources));
-	gatherresourceinfo(ctx, glo, page, rsrc, show);
+	pdf_mark_list_init(ctx, &mark_list);
+	fz_try(ctx)
+		gatherresourceinfo(ctx, &mark_list, glo, page, pageref, show);
+	fz_always(ctx)
+		pdf_mark_list_free(ctx, &mark_list);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
@@ -919,7 +932,10 @@ showinfo(fz_context *ctx, globals *glo, char *filename, int show, const char *pa
 	fz_output *out = glo->out;
 
 	if (!glo->doc)
+	{
 		infousage();
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot show info without document");
+	}
 
 	allpages = !strcmp(pagelist, "1-N");
 
@@ -1022,12 +1038,15 @@ int pdfinfo_main(int argc, char **argv)
 		case 'p': password = fz_optarg; break;
 		default:
 			infousage();
-			break;
+			return 1;
 		}
 	}
 
 	if (fz_optind == argc)
+	{
 		infousage();
+		return 1;
+	}
 
 	ctx = fz_new_context(NULL, NULL, FZ_STORE_UNLIMITED);
 	if (!ctx)
